__author__ = 'aymgal'

from typing import Union

from coolest.template.classes.grid import PixelatedRegularGrid
from coolest.template.classes.noise import Noise
from coolest.template.classes.base import APIBaseObject


class Observation(APIBaseObject):
    """Defines the observation itself, that is the image pixels, 
    the exposure time, the noise model and/or properties, the magnitude 
    zero-point and sky brightness.

    Parameters
    ----------
    pixels : PixelatedRegularGrid, optional
        Regular 2D Grid instance for the observed / mock data pixels, by default None
    noise : Noise, optional
        Instance of a Noise object associated with the modeling 
        of the data pixels, by default None
    mag_zero_point : float, optional
        Zero-point magnitude, which corresponds to the 1 electron per second
        hitting the detector (given in mag), by default None
    mag_sky_brightness : float, optional
        Magnitude due to sky brightness (given in mag per arcsec^2), 
        by default None
    """
    
    def __init__(self, 
                 pixels: PixelatedRegularGrid = None,  # can be None for mock generation
                 exposure_time: Union[float, PixelatedRegularGrid] = None,
                 noise: Noise = None,
                 mag_zero_point: float = None,
                 mag_sky_brightness: float = None,
                 # time_delays: list = None,
                 # magnification_ratios: list = None
                 ) -> None:
        if pixels is None:
            pixels = PixelatedRegularGrid()
        self.pixels = pixels
        self.exposure_time = exposure_time
        self.mag_zero_point = mag_zero_point          # magnitude zero-point (corresponds to 1 electron per second on the detector)
        self.mag_sky_brightness = mag_sky_brightness  # sky brightness (magnitude per arcsec^2)
        if noise is None:
            noise = Noise()
        self.noise = noise
        # self.time_delays = time_delays
        # self.magnification_ratios = magnification_ratios
        super().__init__()

    def check_consistency_with_instrument(self, instrument):
        """Checks that the data image is consistent with instrument properties"""
        width  = abs(self.pixels.field_of_view_x[1]  - self.pixels.field_of_view_x[0])
        height = abs(self.pixels.field_of_view_y[1] - self.pixels.field_of_view_y[0])
        num_pix_ra = int(width / instrument.pixel_size)
        error_message_ra = f"Field-of-view along RA is inconsistent (data: {self.pixels.num_pix_x}, instrument: {num_pix_ra})."
        assert self.pixels.num_pix_x  == num_pix_ra, error_message_ra
        num_pix_dec = int(height / instrument.pixel_size)
        error_message_dec = f"Field-of-view along Dec is inconsistent (data: {self.pixels.num_pix_y}, instrument: {num_pix_dec})."
        assert self.pixels.num_pix_y  == num_pix_dec, error_message_dec
        # TODO: check pixel size value?

    # def set_default_field_of_view(self, instrument):
    #     if not self.pixels.exists:
    #         return
    #     pixel_size = self.pixels.pixel_size
    #     num_pix_x, num_pix_y = self.pixels.shape
    #     self.field_of_view  = ( num_pix_x*pixel_size/2, -num_pix_x*pixel_size/2)  # RA is opposite to x
    #     self.field_of_view_dec = (-num_pix_y*pixel_size/2,  num_pix_y*pixel_size/2)

    # def _check_images(self):
    #     self._check_positive(self.wht_map, "WHT map")
    #     self._check_binary(self.arc_mask, "Arc mask")
    #     self._check_binary(self.likelihood_mask, "Likelihood mask")

    # @staticmethod
    # def _check_positive(fits_file, fits_name):
    #     if not fits_file.exists:
    #         return
    #     pixels, _ = fits_file.read()
    #     if not np.all(pixels > 0):
    #         raise ValueError(f"{fits_name} pixels should all be positive.")

    # @staticmethod
    # def _check_binary(fits_file, fits_name):
    #     if not fits_file.exists:
    #         return
    #     pixels, _ = fits_file.read()
    #     if not np.array_equal(pixels, pixels.astype(bool)):
    #         raise ValueError(f"{fits_name} pixels should be either 0 or 1.")



__author__ = 'aymgal'

from coolest.template.classes.grid import PixelatedRegularGrid
from coolest.template.classes.base import APIBaseObject


__all__ = [
    'Noise',
    'UniformGaussianNoise',
    'NoiseMap',
    'NoiseRealization',
    'InstrumentalNoise',
    'DrizzledNoise',
]

SUPPORTED_CHOICES = list(set(__all__) - {'Noise'})


class Noise(APIBaseObject):
    """Generic class for characterizing the noise of an Observation.
    Supported noise types correspond to the specific classes defined in this module.

    Parameters
    ----------
    ntype : str, optional
        Noise type, as the name of the corresponding class, by default None
    """
    
    def __init__(self, ntype: str = None, **kwargs) -> None:
        self.type = ntype
        for key, value in kwargs.items():
            setattr(self, key, value)
        super().__init__()


class UniformGaussianNoise(Noise):
    """Uniform gaussian noise given a standard deviation and zero mean.

    Parameters
    ----------
    std_dev : float, optional
        Standard deviation (RMS value, since mean = 0), by default 0.
    """
    
    def __init__(self, std_dev: float = 0.) -> None:
        ntype = self.__class__.__name__
        super().__init__(ntype, std_dev=std_dev)


class NoiseMap(Noise):
    """Noise characterized by a noise map, which contains diagonal elements of the data covariance matrix.

    Parameters
    ----------
    noise_map : PixelatedRegularGrid, optional
        Regular 2D Grid instance for the standard deviation of the noise per pixel, by default None
    """

    def __init__(self, noise_map: PixelatedRegularGrid = None) -> None:
        ntype = self.__class__.__name__
        if noise_map is None:
            noise_map = PixelatedRegularGrid()
        super().__init__(ntype, noise_map=noise_map)


class NoiseRealization(Noise):
    """A single realization of the noise.

    Parameters
    ----------
    noise_realization : PixelatedRegularGrid, optional
        Regular 2D Grid instance for the noise value per pixel, by default None
    """
    
    def __init__(self, noise_realization: PixelatedRegularGrid = None) -> None:
        ntype = self.__class__.__name__
        if noise_realization is None:
            noise_realization = PixelatedRegularGrid()
        super().__init__(ntype, noise_realization=noise_realization)


class InstrumentalNoise(Noise):
    """Noise properties are computed directly based on the observed 
    or modeled flux, and on the Instrument (e.g., readout noise) and 
    Observation (e.g., exposure time, sky brightness, etc.) properties.

    Parameters
    ----------
    with_readout_noise : bool, optional
        If True, the noise includes readout noise from the detector, by default True
    with_sky_shot_noise : bool, optional
        If True, the noise includes shot noise from sky background flux 
        (as the Gaussian approximation of the Poisson noise), by default True
    with_target_shot_noise : bool, optional
        If True, the noise includes shot noise from the target flux 
        (as the Gaussian approximation of the Poisson noise), by default True
    """

    def __init__(self, 
                 with_readout_noise: bool = True, 
                 with_sky_shot_noise: bool = True,
                 with_target_shot_noise: bool = True) -> None:
        ntype = self.__class__.__name__
        super().__init__(ntype, 
                         with_readout_noise=with_readout_noise,
                         with_sky_shot_noise=with_sky_shot_noise,
                         with_target_shot_noise=with_target_shot_noise)


class DrizzledNoise(Noise):
    """Provide an exposure map as output by e.g., astrodrizzle with HST images (typically with '_wht' prefix).

    Parameters
    ----------
    background_rms : float, optional
        Noise standard deviation (RMS value, since mean = 0), by default 0.0
    wht_map : PixelatedRegularGrid, optional
        Regular 2D Grid instance for the effective exposure time per pixel, by default None
    """

    def __init__(self, background_rms: float = 0.0, 
                 wht_map: PixelatedRegularGrid = None) -> None:
        ntype = self.__class__.__name__
        if wht_map is None:
            wht_map = PixelatedRegularGrid()
        super().__init__(ntype, background_rms=background_rms, wht_map=wht_map)



__author__ = 'aymgal'

from coolest.template.classes.grid import PixelatedRegularGrid
from coolest.template.classes.base import APIBaseObject


__all__ = [
    'PSF',
    'PixelatedPSF',
    'GaussianPSF',
]

SUPPORTED_CHOICES = list(set(__all__) - {'PSF'})

#TODO: add support for a Moffat PSF


class PSF(APIBaseObject):
    """Defines the model of the Point Spread Function (PSF) of the instrument.

    Parameters
    ----------
    psf_type : str
        Type of PSF model, as the names of the classes defined in this module.
    description : str, optional
        Any details regarding the way the PSF has been estimated, by default None
    **kwargs : dict, optional
        All the other attributes of the specific PSF model.
    """

    def __init__(self,
                 psf_type: str = None,
                 description: str = None,
                 **kwargs) -> None:
        self.type = psf_type
        self.description = description
        for key, value in kwargs.items():
            setattr(self, key, value)
        super().__init__()


class PixelatedPSF(PSF):
    """PSF modeled as a 2D image.

    Parameters
    ----------
    pixels : PixelatedRegularGrid, optional
        Grid class that holds pixel values and the extent of 
        the PSF kernel, by default None
    description : str, optional
        Any details regarding the way the PSF has been estimated, by default None
    """

    def __init__(self, pixels: PixelatedRegularGrid = None, 
                 description: str = None) -> None:
        psf_type = self.__class__.__name__
        if pixels is None:
            pixels = PixelatedRegularGrid()
        super().__init__(psf_type, description=description, pixels=pixels)


class GaussianPSF(PSF):
    """PSF modeled as a 2D image.

    Parameters
    ----------
    fwhm : float, optional
        Full width at half maximum, in units of the Instrument pixel size, 
        by default 0.0
    description : str, optional
        Any details regarding the way the PSF has been estimated, by default None
    """

    def __init__(self, fwhm: float = 0.0, 
                 description: str = None) -> None:
        psf_type = self.__class__.__name__
        super().__init__(psf_type, description=description, fwhm=fwhm)



__author__ = 'aymgal', 'sibirrer'  # partly based on lenstronomy and herculens routines


import numpy as np
import copy

from coolest.api import util


class Coordinates(object):
    """Object that holds the grid of pixel coordinates and provides
    conversions between coordinates in pixel units ('i', 'j') 
    and physical ('x', 'y', in arcseconds) units.

    Parameters
    ----------
    nx : int
        Number of pixels along the x axis
    ny : int
        Number of pixels along the y axis
    matrix_ij_to_xy : 2x2 ndarray
        Matrix such that when multiplied to vector of coordinates in pixel
        units, it returns the vector in physical units.
        Note that it should be diagonal according to COOLEST conventions.
    x_at_ij_0 : physical
        x-coordinate of the pixel with index (0, 0)
    y_at_ij_0 : _type_
        y-coordinate of the pixel with index (0, 0)
    """

    def __init__(self, nx, ny, matrix_ij_to_xy, x_at_ij_0, y_at_ij_0):
        self._matrix_pix2ang = matrix_ij_to_xy
        self._matrix_ang2pix = np.linalg.inv(self._matrix_pix2ang)
        self._ra_at_xy_0 = x_at_ij_0
        self._dec_at_xy_0 = y_at_ij_0
        self._x_at_radec_0, self._y_at_radec_0 \
            = self.map_coord(-self._ra_at_xy_0, -self._dec_at_xy_0, 
                             0, 0, self._matrix_ang2pix)
        self._nx = nx
        self._ny = ny
        self._x_grid, self._y_grid = self.coordinate_grid_2d(nx, ny)
        self._model_grids = {}

    @property
    def pixel_area(self):
        return np.abs(np.linalg.det(self._matrix_pix2ang))

    @property
    def pixel_size(self):
        return np.sqrt(self.pixel_area)

    @property
    def num_points(self):
        return self._nx * self._ny

    @property
    def pixel_coordinates(self):
        return self._x_grid, self._y_grid

    @property
    def pixel_axes(self):
        return self._x_grid[0, :], self._y_grid[:, 0]

    @property
    def extent(self):
        """set of extreme coordinates points"""
        return [self._x_grid[0, 0], self._x_grid[-1, -1], self._y_grid[0, 0], self._y_grid[-1, -1]]

    @property
    def plt_extent(self):
        """set of coordinates of the borders of the grid (useful for matplotlib functions)"""
        extent = copy.copy(self.extent)
        # WARNING: the following assumes NO ROTATION (i.e. coordinates axes aligned with x/y axes)
        half_pix = self.pixel_size / 2.
        pix_scl_x = self._matrix_pix2ang[0, 0]
        pix_scl_y = self._matrix_pix2ang[1, 1]
        if self.x_is_inverted:
            extent[0] += pix_scl_x / 2.
            extent[1] -= pix_scl_x / 2.
        else:
            extent[0] -= pix_scl_x / 2.
            extent[1] += pix_scl_x / 2.
        if self.y_is_inverted:
            extent[2] += pix_scl_y / 2.
            extent[3] -= pix_scl_y / 2.
        else:
            extent[2] -= pix_scl_y / 2.
            extent[3] += pix_scl_y / 2.
        return extent

    @property
    def shape(self):
        return self._nx * self.pixel_size, self._ny * self.pixel_size

    @property
    def center(self):
        return np.mean(self._x_grid), np.mean(self._y_grid)

    @property
    def x_is_inverted(self):
        return self._matrix_pix2ang[0, 0] < 0

    @property
    def y_is_inverted(self):
        return self._matrix_pix2ang[1, 1] < 0

    @staticmethod
    def map_coord(ra, dec, x_0, y_0, M):
        x, y = np.array(M).dot(np.array([ra, dec]))
        return x + x_0, y + y_0

    def radec_to_pixel(self, ra, dec):
        return self.map_coord(ra, dec, self._x_at_radec_0, self._y_at_radec_0, self._matrix_ang2pix)

    def pixel_to_radec(self, x, y):
        return self.map_coord(x, y, self._ra_at_xy_0, self._dec_at_xy_0, self._matrix_pix2ang)

    @staticmethod
    def grid_from_coordinate_transform(nx, ny, Mpix2coord, x_at_ij_0, y_at_ij_0):
        a = np.arange(nx)
        b = np.arange(ny)
        matrix = np.dstack(np.meshgrid(a, b)).reshape(-1, 2)
        x_grid = matrix[:, 0]
        y_grid = matrix[:, 1]
        ra_grid = x_grid * Mpix2coord[0, 0] + y_grid * Mpix2coord[0, 1] + x_at_ij_0
        dec_grid = x_grid * Mpix2coord[1, 0] + y_grid * Mpix2coord[1, 1] + y_at_ij_0
        return ra_grid, dec_grid

    def coordinate_grid_1d(self, nx, ny):
        ra_coords, dec_coords = self.grid_from_coordinate_transform(nx, ny, self._matrix_pix2ang, self._ra_at_xy_0, self._dec_at_xy_0)
        return ra_coords, dec_coords

    def coordinate_grid_2d(self, nx, ny):
        ra_coords, dec_coords = self.coordinate_grid_1d(nx, ny)
        ra_coords = util.array2image(ra_coords, nx, ny)
        dec_coords = util.array2image(dec_coords, nx, ny)
        return ra_coords, dec_coords

    def create_new_coordinates(self, pixel_scale_factor=None, 
                               grid_center=None, grid_shape=None):
        """Based on the current coordinates, creates a sub-coordinates system
        potentially shifted, with a different pixel size and field-of-view.

        Parameters
        ----------
        pixel_scale_factor : float, optional
            Ratio between the new pixel size and the original pixel size, by default None
        grid_center : (float, float), optional
            x and y center coordinates of the new coordinates grid, by default None
        grid_shape : (float, float), optional
            Width and height (in arcsec) of the new coordinates grid, by default None

        Returns
        -------
        Coordinates
            New instance of a Coordinates object
        """
        unchanged_count = 0
        if grid_center is None or grid_center == self.center:
            grid_center_ = self.center
            unchanged_count += 1
        else:
            grid_center_ = grid_center
        if grid_shape is None or grid_shape == self.shape:
            grid_shape_ = self.shape
            unchanged_count += 1
        else:
            grid_shape_ = grid_shape
        if pixel_scale_factor is None or pixel_scale_factor == 1:
            pixel_scale_factor_ = 1
            unchanged_count += 1
        else:
            pixel_scale_factor_ = pixel_scale_factor

        # in case it's the same region as the base coordinate grid
        if unchanged_count == 3:
            return copy.deepcopy(self)

        pixel_size = self.pixel_size * float(pixel_scale_factor_)
        center_x, center_y = grid_center_
        width, height = grid_shape_
        nx = round(width / pixel_size)
        ny = round(height / pixel_size)

        matrix_pix2ang = self._matrix_pix2ang / self.pixel_size * pixel_size

        cx, cy = nx / 2., ny / 2.
        cra, cdec = matrix_pix2ang.dot(np.array([cx, cy]))
        x_at_ij_0, y_at_ij_0 = - cra + center_x + pixel_size/2., - cdec + center_y + pixel_size/2.

        return Coordinates(nx, ny, matrix_pix2ang, x_at_ij_0, y_at_ij_0)



__author__ = 'aymgal'


import os
from astropy.io import fits


class FitsFile(object):
    """Class that represents a FITS file on the disk.

    Parameters
    ----------
    path : str
        Path to the FITS file, depending on usage either relative or absolute.
    directory : str, optional
        Absolute directory containing the FITS file, 
        in case `path` is a relative, by default None
    check_exist : bool, optional
        If True, will check if the FITS file exists when the object is 
        instantiated, by default False

    Raises
    ------
    RuntimeError
        If the FITS file does not exist on the disk.
    """

    def __init__(self, path: str, 
                 directory: str = None, 
                 check_exist: bool = False) -> None:
        self.path = path
        self._directory = directory
        if not self.exists() and check_exist:
            raise RuntimeError(f"FITS file located at '{self.path}' does not exist")

    @property
    def abs_path(self):
        if not hasattr(self, '_directory') or self._directory is None:
            return self.path
        return os.path.join(self._directory, self.path)

    def exists(self):
        if self.path is None:
            return False
        return os.path.exists(self.abs_path)

    def read(self, directory=None):
        """Read the data and header content of the FITS file, using astropy.io.fits.

        A directory must be given typically when it has not been set at 
        the initialization of the object.

        Parameters
        ----------
        directory : str, optional
            Absolute directory containing the FITS file, by default None

        Returns
        -------
        ndarray or recarray or Group
            (data, header) content of the FITS file (as returned by fits.getdata())

        Raises
        ------
        ValueError
            If no directory is provided and no directory has been set beforehand.
        """
        if not hasattr(self, '_directory'):
            if directory is None:
                raise ValueError("You must provide a FITS file directory if none has been set")
            abs_path = os.path.join(directory, self.path)
        else:
            abs_path = self.abs_path
        return fits.getdata(abs_path, header=True)



__author__ = 'aymgal'

from coolest.template.classes.base import APIBaseObject


class Cosmology(APIBaseObject):
    """Defines the cosmological model. 
    Currently, only FlatLambdaCDM from astropy is supported, based on H0 and Omega_m.

    Parameters
    ----------
    H0 : float
        Hubble constant, in km/s/Mpc
    Om0 : float
        Matter density at present time
    astropy_name : str, optional
        `astropy` model name, by default 'FlatLambdaCDM'
    """
    
    def __init__(self, 
                 H0: float,
                 Om0: float,
                 astropy_name: str = 'FlatLambdaCDM') -> None:
        self.H0 = H0
        self.Om0 = Om0
        self.astropy_name = astropy_name
        super().__init__()



__author__ = 'aymgal'


import os
import math
import numpy as np
# from astropy.coordinates import SkyCoord
from skimage import measure

from coolest.template.json import JSONSerializer


def convert_image_to_data_units(image, mag_tot, mag_zero_point=None, coolest_object=None):
    """
    Rescale an image such that it has units of electrons per second (e/s),
    given a total magnitude and a magnitude zero-point.

    After rescaling, the total magnitude of the image_rescaled should corresponds to

    `-2.5 * np.log10(image_rescaled.sum()) + mag_zero_point = mag_tot`

    where `mag_zero_point` corresponds to the magnitude of 1 e/s.
    :param image: input image, as a 2D array.
    :param mag_tot: target total magnitude, integrated over the whole image
    :param mag_zero_point: magnitude zero point of the observation (that corresponds to 1 e/s). If coolest_object is not None, mag_zero_point is ignored.
    :param coolest_object: if given, will be used to retrieve the zero-point magnitude of the observation.
    """
    if coolest_object is None and mag_zero_point is None:
        raise ValueError("Either a COOLEST object or a zero-point magnitude should be provided.")
    flux_tot = np.sum(image)
    image_unit_flux = image / flux_tot
    if coolest_object is not None:
        mag_zp = coolest.observation.mag_zero_point
    else:
        mag_zp = mag_zero_point
    delta_mag = mag_tot - mag_zp
    flux_unit_mag = 10 ** ( - delta_mag / 2.5 )
    image_rescaled = image_unit_flux * flux_unit_mag
    return image_rescaled


def get_coolest_object(file_path, verbose=False, **kwargs_serializer):
    if not os.path.isabs(file_path):
        file_path = os.path.abspath(file_path)
    serializer = JSONSerializer(file_path, **kwargs_serializer)
    return serializer.load(verbose=verbose)


def get_coordinates(coolest_object, offset_x=0., offset_y=0.):
    from coolest.api.coordinates import Coordinates  # prevents circular import errors
    nx, ny = coolest_object.observation.pixels.shape
    pix_scl = coolest_object.instrument.pixel_size
    half_size_x, half_size_y = nx * pix_scl / 2., ny * pix_scl / 2.
    x_at_ij_0  = - half_size_x + pix_scl / 2.  # position of x=0 with respect to bottom left pixel
    y_at_ij_0 = - half_size_y + pix_scl / 2.  # position of y=0 with respect to bottom left pixel
    matrix_pix2ang = pix_scl * np.eye(2)  # transformation matrix pixel <-> angle
    return Coordinates(nx, ny, matrix_ij_to_xy=matrix_pix2ang,
                       x_at_ij_0=x_at_ij_0 + offset_x, 
                       y_at_ij_0=y_at_ij_0 + offset_y)


def get_coordinates_from_regular_grid(field_of_view_x, field_of_view_y, num_pix_x, num_pix_y):
    from coolest.api.coordinates import Coordinates  # prevents circular import errors
    pix_scl_x = np.abs(field_of_view_x[0] - field_of_view_x[1]) / num_pix_x
    pix_scl_y = np.abs(field_of_view_y[0] - field_of_view_y[1]) / num_pix_y
    matrix_pix2ang = np.array([[pix_scl_x, 0.], [0., pix_scl_y]])
    x_at_ij_0  = field_of_view_x[0] + pix_scl_x / 2.
    y_at_ij_0 = field_of_view_y[0] + pix_scl_y / 2.
    return Coordinates(
        num_pix_x, num_pix_y, matrix_ij_to_xy=matrix_pix2ang,
        x_at_ij_0=x_at_ij_0, y_at_ij_0=y_at_ij_0,
    )


def get_coordinates_set(coolest_file_list, reference=0):
    coordinates_list = []
    for coolest_file in coolest_file_list:

        # TODO: compute correct offsets when each file has
        # obs = self.coolest.observation
        # sky_coord = SkyCoord(obs.ra, obs.dec, frame='icrs')
        # ra, dec = sky_coord.to_string(style='hmsdms').split(' ')

        coordinates = get_coordinates(coolest_file)
        coordinates_list.append(coordinates)
    return coordinates_list


def array2image(array, nx=0, ny=0):
    """Convert a 1d array into a 2d array.

    Note: this only works when length of array is a perfect square, or else if
    nx and ny are provided

    :param array: image values
    :type array: array of size n**2
    :returns:  2d array
    :raises: AttributeError, KeyError
    """
    if nx == 0 or ny == 0:
        # Avoid turning n into a JAX-traced object with jax.numpy.sqrt
        n = int(np.sqrt(len(array)))
        if n**2 != len(array):
            err_msg = f"Input array size {len(array)} is not a perfect square."
            raise ValueError(err_msg)
        nx, ny = n, n
    image = array.reshape(int(nx), int(ny))
    return image


def image2array(image):
    """Convert a 2d array into a 1d array.

    :param array: image values
    :type array: array of size (n,n)
    :returns:  1d array
    :raises: AttributeError, KeyError
    """
    # nx, ny = image.shape  # find the size of the array
    # imgh = np.reshape(image, nx * ny)  # change the shape to be 1d
    # return imgh
    return image.ravel()


def downsampling(image, factor=1):
    if factor < 1:
        raise ValueError(f"Downscaling factor must be > 1")
    if factor == 1:
        return image
    f = int(factor)
    nx, ny = np.shape(image)
    if int(nx/f) == nx/f and int(ny/f) == ny/f:
        down = image.reshape([int(nx/f), f, int(ny/f), f]).mean(3).mean(1)
        return down
    else:
        raise ValueError(f"Downscaling factor {factor} is not possible with shape ({nx}, {ny})")


def effective_radius(light_map, x, y, outer_radius=10, initial_guess=1, initial_delta_pix=10, n_iter=10):
    """Computes the effective radius of the 2D surface brightness profile, 
    based on a definition similar to the half-light radius.
    NOTE: This functions assumes that the profile is centered on the grid.

    Parameters
    ----------
    light_map : ndarray
        2D array of the light model
    x : ndarray
        x-coordinates associated to the light model
    y : ndarray
        y-coordinates associated to the light model
    outer_radius : int, optional
        outer limit of integration within which half the light is calculated to estimate the effective radius, by default 10
    initial_guess : int, optional
        Initial guess for effective radius in arcsecond, by default 1
    initial_delta_pix : int, optional
        Initial step size in pixels before shrinking in future iterations, by default 10
    n_iter : int, optional
        Number of iterations, by default 5

    Returns
    -------
    (float, float)
        Effective radius and spacing of the coordinates grid (approximate accuracy)

    Raises
    ------
    RuntimeError
        If integration loop exceeds outer bound before convergence.
    """
    #initialize
    grid_res=np.abs(x[0,0]-x[0,1])
    initial_delta=grid_res*initial_delta_pix #default inital step size is 10 pixels
    r_grid=np.hypot(x, y)
    total_light=np.sum(light_map[r_grid<outer_radius])
    cumulative_light=np.sum(light_map[r_grid<initial_guess])
    if cumulative_light < total_light/2: #move outward
        direction=1
    elif cumulative_light > total_light/2: #move inward
        direction=-1
    else:
        return initial_guess
    r_eff=initial_guess
    delta=initial_delta
    loopcount=0

    for n in range(n_iter): #overshoots, turn around and backtrack at higher precision
        while (total_light/2.-cumulative_light)*direction>0: 
            if loopcount > 100:
                raise RuntimeError('Stuck in very long (possibly infinite) loop')
            r_eff=r_eff+delta*direction
            cumulative_light=np.sum(light_map[r_grid<r_eff])
            loopcount+=1
        direction=direction*-1
        delta=delta/2.

    return r_eff, grid_res
            

def ellipticity_from_moments(light_map, pixel_size):
    # compute central momoments
    try:
        # scikit-image version 0.19.3 and older
        mu = measure.moments_central(light_map, order=2, spacing=(pixel_size, pixel_size))
    except IndexError:
        # scikit-image version 0.20.0 and beyond
        mu = measure.moments_central(light_map, order=2, spacing=pixel_size)

    # use the moments to estimate orientation and ellipticity (https://en.wikipedia.org/wiki/Image_moment)
    mu_20_ = mu[2, 0] / mu[0, 0]
    mu_02_ = mu[0, 2] / mu[0, 0]
    mu_11_ = mu[1, 1] / mu[0, 0]
    lambda_1 = (mu_20_ + mu_02_) / 2. + np.sqrt(4*mu_11_**2 + (mu_20_ - mu_02_)**2) / 2.
    lambda_2 = (mu_20_ + mu_02_) / 2. - np.sqrt(4*mu_11_**2 + (mu_20_ - mu_02_)**2) / 2.
    q = np.sqrt(lambda_2 / lambda_1)  # b/a, axis ratio
    phi = np.arctan(2. * mu_11_ / (mu_20_ - mu_02_)) / 2.  # position angle
    if mu_02_ > mu_20_:
        phi += np.pi / 2.  # makes it consistent angles conventions in COOLEST
    return phi, q


def azim_averaged_two_point_correlation(light_image, dpix, rmax, Nbins):
    """
    The two point correlation function can be obtained from the covariance matrix of an image and the distances between its pixels.
    By binning the covariance matrix entries in distance (or radial) bins, one can obtain the 1D correlation function.
    There are two ways to obtain the covariance matrix:
    1) it is equivalent to the inverse Fourier transform of the power spectrum, and
    2) by calculating explicitly the covariance between any two pixels
    Here we use the first way.

    Parameters
    ----------
    light_image : 2D ndarray
        Pixels of the image to analyse.
    dpix : float
        Pixel size
    Nbins : int, optional
        The number of radial bins to use for converting the 2D covariance matrix into a 1D correlation function.
    rmax : float, optional
        A value for the maximum extent of the radial bins. If none is given then it is equal to half the diagonal of the provided image. 

    Returns
    -------
    (array, array, array, array)
        The location, value, uncertainty and covariance matrix 
        The covariance matrix here is the inverse of fourier transform of power spectrum)
    """
    # Fourier transform image
    fouriertf = np.fft.fft2(light_image, norm='ortho')
    # Power spectrum (the square of the signal)
    absval2 = fouriertf.real**2 + fouriertf.imag**2
    # Covariance matrix (the inverse fourier transform of the power spectrum)
    complex_cov = np.fft.fftshift(np.fft.ifft2(absval2, norm='ortho'))
    cov = complex_cov.real
    
    # Bin the 2D covariance matrix into radial bins
    rmin = 0.0
    dr = (rmax-rmin)/Nbins
    bins = np.arange(rmin,rmax,dr)
    vals = [[] for _ in range(len(bins))]
    # Ni = Nj = the total number of pixels in the image
    Ni = cov.shape[1]
    Nj = cov.shape[0]
    for i in range(0,Ni):
        for j in range(0,Nj):
            r = np.hypot((j-Nj/2.0)*dpix,(i-Ni/2.0)*dpix)
            if r < rmax and i != j:
                index = int(np.floor(r/dr))
                vals[index].append(cov[i][j])
                
    means = np.zeros(len(bins))                    
    sdevs = np.zeros(len(bins))
    for i in range(0,len(bins)):
        if len(vals[i]) > 0:
            means[i] = np.mean(vals[i])
            sdevs[i] = np.std(vals[i])

    return bins, means, sdevs, cov
    

def lensing_information(data_lens_sub, x, y, theta_E, noise_map, center_x_lens=0, center_y_lens=0,
                        a=16, b=0, arc_mask=None):
    """
    Computes the 'lensing information' defined in Tan et al. 2023, Equations (8) and (9).
    https://ui.adsabs.harvard.edu/abs/2023arXiv231109307T/abstract

    Parameters
    ----------
    data_lens_sub : np.ndarray
        Imaging data as a 2D array. It is assumed to contain no lens light.
    x : np.ndarray
        2D array of x coordinates, in arcsec.
    y : np.ndarray
        2D array of y coordinates, in arcsec.
    theta_E : float
        Einstein radius in arcsec, by default None
    noise_map : np.ndarray
        2D array with 1-sigma noise level per pixel (same units as `data_lens_sub`), by default None
    center_x_lens : int, optional
        x coordinates of the center of the lens, by default 0
    center_y_lens : int, optional
        y coordinates of the center of the lens, by default 0
    a : int, optional
        Exponent in Eq. (9) from Yi Tan et al. 2023, by default 16
    b : int, optional
        Exponent in Eq. (9) from Yi Tan et al. 2023, by default 0
    arc_mask : np.ndarray, optional
        Binary 2D array with 1s where there is are lensed arcs, by default None

    Returns
    -------
    4-tuple
        Lensing information I, Einstein radius, reference azimuthal angle, total mask used for computing I
    """
    if arc_mask is None:
        arc_mask = np.ones_like(data_lens_sub)
    # estimate background noise from one corner of the noise map
    sigma_bkg = np.mean(noise_map[:10, :10])
    # build a mask to only consider pixels at least 3 times the background noise level
    snr_mask = np.where(data_lens_sub > 3.*sigma_bkg, 1., 0.)
    # combine user mask and SNR mask
    arc_mask_tot = snr_mask * arc_mask
    # shift coordinates so that lens is at (0, 0)
    theta_x, theta_y = x - center_x_lens, y - center_y_lens
    # compute polar coordinates centered on the lens
    theta_r = np.hypot(theta_x, theta_y)
    phi = np.arctan2(theta_y, theta_x)
    # find index of the brightest pixel (within the arc mask)
    max_idx = np.where(data_lens_sub == (data_lens_sub*arc_mask_tot).max())
    # get azimuthal angle corresponding to the brightest pixel
    phi_ref = float(np.arctan2(theta_y[max_idx], theta_x[max_idx]))
    # compute the weights following Eq. (9) from Yi Tan et al. 2023
    weights = ( 1. + np.abs(theta_r - theta_E) / theta_E * (1 + np.abs(phi - phi_ref) / phi_ref)**b )**a
    # compute the weighted sum
    numerator = np.sum(arc_mask_tot*weights*data_lens_sub)
    denominator = np.sqrt(np.sum(arc_mask_tot*noise_map**2))
    lens_I = numerator / denominator
    return lens_I, theta_E, phi_ref, arc_mask_tot


def split_lens_source_params(coolest_list, name_list, lens_light=False):
    """
    Read several json files already containing a model with the results of this model fitting

    INPUT
    -----
    file_list: list, list of path or names of the file to read
    name_list: list, list of shorter names to distinguish the files
    lens_light: bool, if True, computes the lens light kwargs as well (not yet implemented)

    OUTPUT
    ------
     param_all_lens, param_all_source: organized dictionaries readable by plotting function
    """

    param_all_lens = {}
    param_all_source = {}
    for idx_file, file_name in enumerate(name_list):

        print(file_name)
        coolest_obj = coolest_list[idx_file]

        if coolest_obj.mode == 'MAP':
            print('LENS COOLEST : ', coolest_obj.mode)
        else:
            print('LENS COOLEST IS NOT MAP, BUT IS ', coolest_obj.mode)

        lensing_entities_list = coolest_obj.lensing_entities

        param_lens = {}
        param_source = {}

        if lensing_entities_list is not None:

            creation_lens_source_light = True

            idx_lens = 0
            idx_lens_light = 0
            idx_source = 0
            idx_ps = 0

            min_red = 0
            max_red = 5
            creation_red_list = True
            red_list = []
            MultiPlane = False
            for lensing_entity in lensing_entities_list:
                red_list.append(lensing_entity.redshift)
            min_red = np.min(red_list)
            max_red = np.max(red_list)

            for lensing_entity in lensing_entities_list:
                if lensing_entity.type == "Galaxy":
                    galac = lensing_entity

                    if galac.redshift > min_red:
                        # SOURCE OF LIGHT
                        light_list = galac.light_model
                        for light in light_list:

                            if light.type == 'Sersic':
                                read_sersic(light, param_source)
                                idx_source += 1
                            else:
                                print('Light Type ', light.type, ' not yet implemented.')

                    if galac.redshift < max_red:
                        # LENSING GALAXY
                        if galac.redshift > min_red:
                            MultiPlane = True
                            print('Multiplane lensing to consider.')
                        mass_list = galac.mass_model
                        for mass in mass_list:

                            if mass.type == 'PEMD':
                                read_pemd(mass, param_lens)
                                idx_lens += 1
                            elif mass.type == 'SIE':
                                read_sie(mass, param_lens)
                                idx_lens += 1
                            else:
                                print('Mass Type ', mass.type, ' not yet implemented.')

                    if (galac.redshift <= min_red) and (galac.redshift >= max_red):
                        print('REDSHIFT ', galac.redshift, ' is not in the range ]', min_red, ',', max_red, '[')

                elif lensing_entity.type == "MassField":
                    shear_list = lensing_entity.mass_model
                    for shear_idx in shear_list:

                        if shear_idx.type == 'ExternalShear':
                            read_shear(shear_idx, param_lens)
                            idx_lens += 1
                        else:
                            print("type of Shear ", shear_idx.type, " not implemented")
                else:
                    print("Lensing entity of type ", lensing_enity.type, " is unknown.")

        param_all_lens[name_list[idx_file]] = param_lens
        param_all_source[name_list[idx_file]] = param_source

    return param_all_lens, param_all_source


def read_shear(mass, param={}, prefix='SHEAR_0_'):
    """
    Reads the parameters of a coolest.template.classes.profiles.mass.ExternalShear object

    INPUT
    -----
    mass : coolest.template.classes.profiles.mass.ExternalShear object
    param : dict, already existing dictionnary with ordered parameters readable by plotting function
    prefix : str, prefix to use in saving parameters names

    OUTPUT
    ------
    param : updated param
    """

    for mass_name, mass_param in mass.parameters.items():
        p = getattr(mass_param.point_estimate, 'value')
        p_84 = getattr(mass_param.posterior_stats, 'percentile_84th')
        p_16 = getattr(mass_param.posterior_stats, 'percentile_16th')
        p_med = getattr(mass_param.posterior_stats, 'median')
        p_mean = getattr(mass_param.posterior_stats, 'mean')
        latex_name = getattr(mass_param, 'latex_str')
        if mass_name == 'gamma_ext':
            param[prefix + 'gamma_ext'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                           'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif mass_name == 'phi_ext':
            param[prefix + 'phi_ext'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                         'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        else:
            print(shear_name, " not known")
    print('\t Shear correctly added')

    return param


def read_pemd(mass, param={}, prefix='PEMD_0_'):
    """
    Reads the parameters of a coolest.template.classes.profiles.mass.PEMD object

    INPUT
    -----
    mass : coolest.template.classes.profiles.mass.PEMD object
    param : dict, already existing dictionnary with ordered parameters readable by plotting function
    prefix : str, prefix to use in saving parameters names

    OUTPUT
    ------
    param : updated param
    """

    for mass_name, mass_param in mass.parameters.items():
        p = getattr(mass_param.point_estimate, 'value')
        p_84 = getattr(mass_param.posterior_stats, 'percentile_84th')
        p_16 = getattr(mass_param.posterior_stats, 'percentile_16th')
        p_med = getattr(mass_param.posterior_stats, 'median')
        p_mean = getattr(mass_param.posterior_stats, 'mean')
        latex_name = getattr(mass_param, 'latex_str')
        if mass_name == 'theta_E':
            param[prefix + 'theta_E'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                         'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif mass_name == 'q':
            param[prefix + 'q'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                   'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif mass_name == 'phi':
            param[prefix + 'phi'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                     'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif mass_name == 'center_x':
            param[prefix + 'cx'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                    'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif mass_name == 'center_y':
            param[prefix + 'cy'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                    'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif mass_name == 'gamma':
            param[prefix + 'gamma'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                       'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        else:
            print(mass_name, " not known")

    print('\t PEMD correctly added')

    return param


def read_sie(mass, param={}, prefix='SIE_0_'):
    """
    Reads the parameters of a coolest.template.classes.profiles.mass.SIE object

    INPUT
    -----
    mass : coolest.template.classes.profiles.mass.SIE object
    param : dict, already existing dictionnary with ordered parameters readable by plotting function
    prefix : str, prefix to use in saving parameters names

    OUTPUT
    ------
    param : updated param
    """

    for mass_name, mass_param in mass.parameters.items():
        p = getattr(mass_param.point_estimate, 'value')
        p_84 = getattr(mass_param.posterior_stats, 'percentile_84th')
        p_16 = getattr(mass_param.posterior_stats, 'percentile_16th')
        p_med = getattr(mass_param.posterior_stats, 'median')
        p_mean = getattr(mass_param.posterior_stats, 'mean')
        latex_name = getattr(mass_param, 'latex_str')
        if mass_name == 'theta_E':
            param[prefix + 'theta_E'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                         'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif mass_name == 'q':
            param[prefix + 'q'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                   'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif mass_name == 'phi':
            param[prefix + 'phi'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                     'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif mass_name == 'center_x':
            param[prefix + 'cx'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                    'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif mass_name == 'center_y':
            param[prefix + 'cy'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                    'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        else:
            print(mass_name, " not known")

    print('\t SIE correctly added')

    return param


def read_sersic(light, param={}, prefix='Sersic_0_'):
    """
    Reads the parameters of a coolest.template.classes.profiles.light.Sersic object

    INPUT
    -----
    mass : coolest.template.classes.profiles.light.Sersic object
    param : dict, already existing dictionnary with ordered parameters readable by plotting function
    prefix : str, prefix to use in saving parameters names

    OUTPUT
    ------
    param : updated param
    """

    for light_name, light_param in light.parameters.items():
        p = getattr(light_param.point_estimate, 'value')
        p_84 = getattr(light_param.posterior_stats, 'percentile_84th')
        p_16 = getattr(light_param.posterior_stats, 'percentile_16th')
        p_med = getattr(light_param.posterior_stats, 'median')
        p_mean = getattr(light_param.posterior_stats, 'mean')
        latex_name = getattr(light_param, 'latex_str')
        if light_name == 'I_eff':
            param[prefix + 'A'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                   'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif light_name == 'n':
            param[prefix + 'n_sersic'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                          'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif light_name == 'theta_eff':
            param[prefix + 'R_sersic'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                          'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif light_name == 'q':
            param[prefix + 'q'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                   'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif light_name == 'phi':
            param[prefix + 'phi'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                     'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif light_name == 'center_x':
            param[prefix + 'cx'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                    'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        elif light_name == 'center_y':
            param[prefix + 'cy'] = {'point_estimate': p, 'percentile_84th': p_84, 'percentile_16th': p_16,
                                    'median': p_med, 'mean': p_mean, 'latex_str': latex_name}
        else:
            print(light_name, " not known")

    print('\t Sersic correctly added')

    return param


def find_critical_lines(coordinates, mag_map):
    # invert and find contours corresponding to infinite magnification (i.e., changing sign)
    inv_mag = 1. / np.array(mag_map)
    contours = measure.find_contours(inv_mag, 0.)
    # convert to model coordinates
    lines = []
    for contour in contours:
        curve_x, curve_y = coordinates.pixel_to_radec(contour[:, 1], contour[:, 0])
        lines.append((np.array(curve_x), np.array(curve_y)))
    return lines


def find_caustics(crit_lines, composable_lens):
    """`composable_lens` can be an instance of `ComposableLens` or `ComposableMass`"""
    lines = []
    for cline in crit_lines:
        cl_src_x, cl_src_y = composable_lens.ray_shooting(cline[0], cline[1])
        lines.append((np.array(cl_src_x), np.array(cl_src_y)))
    return lines


def find_all_lens_lines(coordinates, composable_lens):
    """`composable_lens` can be an instance of `ComposableLens` or `ComposableMass`"""
    from coolest.api.composable_models import ComposableLensModel, ComposableMassModel  # avoiding circular imports 
    if isinstance(composable_lens, ComposableLensModel):
        mag_fn = composable_lens.lens_mass.evaluate_magnification
    elif isinstance(composable_lens, ComposableMassModel):
        mag_fn = composable_lens.evaluate_magnification
    else:
        raise ValueError("`composable_lens` must be a ComposableLensModel or a ComposableMassModel.")
    mag_map = mag_fn(*coordinates.pixel_coordinates)
    crit_lines = find_critical_lines(coordinates, mag_map)
    caustics = find_caustics(crit_lines, composable_lens)
    return crit_lines, caustics
    

def resample_multivariate_normal(samples, num_samples=5_000, **kwargs_cov):
    """Resample following multi-variate normal distribution"""
    mean = np.mean(samples, axis=0)
    cov = np.cov(samples.T, **kwargs_cov)
    num_params = samples.shape[1]
    resampled = np.random.multivariate_normal(
        mean=mean, cov=cov, size=(int(num_samples/num_params), num_params)).reshape((-1, num_params))
    return resampled



__author__ = 'aymgal'

from typing import Tuple

from coolest.template.classes.base import APIBaseObject
from coolest.template.classes.regularizations import pixelated as regularization_module
from coolest.template.classes import util


class RegularizationList(list, APIBaseObject):

    def __init__(self, 
                 *regularization_profile_pairs: Tuple[Tuple]) -> None:
        """
        regularization_profile_pairs is a tuple of tuple with the following syntax:
        (
            ('regularization method 1', specific Profile (light or mass) instance),
            ('regularization method 2', specific Profile (light or mass) instance),
            ...
        )
        """
        profiles = []
        for name, profile in regularization_profile_pairs:
            if name not in regularization_module.SUPPORTED_CHOICES:
                raise ValueError(f"Regularization '{name}' is not supported.")
            profile_class = getattr(regularization_module, name)
            profiles.append(profile_class(applied_to_profile_id=profile.id))
        list.__init__(self, profiles)
        APIBaseObject.__init__(self)
        self._create_all_ids()

    def _create_all_ids(self):
        for j, regularization in enumerate(self):
            regularization_id = util.regul_to_id(regularization.type, j)
            regularization.id = regularization_id
            for name, parameter in regularization.parameters.items():
                param_id = f'{regularization.id}_{name}'
                parameter.id = param_id



__author__ = 'aymgal'

from typing import Dict

from coolest.template.classes.base import APIBaseObject
from coolest.template.classes.parameter import Parameter


class Regularization(APIBaseObject):

    def __init__(self,
                 documentation: str, 
                 parameters: Dict[str, Parameter],
                 applied_to_profile_id: str = None) -> None:
        self.type = self.__class__.__name__  # name of children class
        self.documentation = documentation
        self.parameters = parameters
        self.applied_to_profile_id = applied_to_profile_id
        super().__init__()



__author__ = 'aymgal'


import numpy as np
from scipy import ndimage



def degree2radians(phi):
    return phi * np.pi / 180.

def eastofnorth2normalradians(phi_in):
    phi_out = degree2radians(phi_in - 90.)
    return phi_out

def shift(x, y, center_x, center_y):
    x_shift = x - center_x
    y_shift = y - center_y
    return x_shift, y_shift

def rotate(x, y, phi_radians):
    cos_phi = np.cos(phi_radians)
    sin_phi = np.sin(phi_radians)
    x_trans = cos_phi * x + sin_phi * y
    y_trans = -sin_phi * x + cos_phi * y
    return x_trans, y_trans

def shift_rotate_elliptical(x, y, phi_radians, q, center_x, center_y):
    x_shift, y_shift = shift(x, y, center_x, center_y)
    x_trans, y_trans = rotate(x_shift, y_shift, phi_radians)
    return x_trans * np.sqrt(q), y_trans / np.sqrt(q)

def cartesian2polar(x, y):
    r = np.hypot(x, y)
    phi = np.arctan2(y, x)
    return r, phi


class CartesianGridInterpolator(object):
    """
    Regular grid spline interpolator
    https://docs.scipy.org/doc/scipy/tutorial/interpolate/ND_regular_grid.html#uniformly-spaced-data
    """
    
    def __init__(self, points, values, method='linear', fill_value=0.):
        self.limits = np.array([[min(x), max(x)] for x in points])
        self.values = np.asarray(values, dtype=float)
        self.order = {'linear': 1, 'cubic': 3, 'quintic': 5}[method]
        self.fill_value = fill_value

    def __call__(self, xi):
        """
        `xi` here is an array-like (an array or a list) of points.

        Each "point" is an ndim-dimensional array_like, representing
        the coordinates of a point in ndim-dimensional space.
        """
        # transpose the xi array into the ``map_coordinates`` convention
        # which takes coordinates of a point along columns of a 2D array.
        xi = np.asarray(xi).T

        # convert from data coordinates to pixel coordinates
        ns = self.values.shape
        coords = [(n-1)*(val - lo) / (hi - lo) 
                  for val, n, (lo, hi) in zip(xi, ns, self.limits)]

        # interpolate
        return ndimage.map_coordinates(self.values, coords,
                                       order=self.order,
                                       mode='constant',
                                       cval=self.fill_value,
                                       prefilter=False)



__author__ = 'aymgal', 'gvernard'

from typing import Tuple
import numpy as np
import warnings

from coolest.template.classes.base import APIBaseObject
from coolest.template.classes.fits_file import FitsFile


__all__ = [
    'Grid',
    'PixelatedRegularGrid',
    'IrregularGrid',
]

SUPPORTED_CHOICES = [
    'PixelatedRegularGrid',
    'IrregularGrid',
]


class Grid(APIBaseObject):
    """Generic class that represents a grid of coordinates and values.

    Parameters
    ----------
    fits_path : str
        Path to the FITS file in which the values (and perhaps the coordinates)
        are stored. If meant to be stored within a COOLEST template file, 
        this should just be the name of the FITS file, placed 
        in the same directory as the template.
    check_fits_file : bool, optional
        If True, creating the object will check that the FITS file exists, by default False
    fits_file_dir : str, optional
        Absolute path of the directory containing the FITS file, by default None
    """

    def __init__(self,
                 fits_path: str,
                 check_fits_file: bool = False, 
                 fits_file_dir: str = None) -> None:
        self.fits_file = FitsFile(fits_path, 
                                  check_exist=check_fits_file, 
                                  directory=fits_file_dir)
        super().__init__()

    def set_grid(self, fits_path, check_fits_file):
        """Set / replace the FitsFile object associated with the Grid.
        This is useful to set associate the FITS file after the Grid instance
        has been created.

        Parameters
        ----------
        fits_path : str
            Path to the FITS file
        check_fits_file : _type_
            If True, check the existence of the FITS file at fits_path.
        """
        if fits_path is not None:
            self.fits_file = FitsFile(fits_path, check_exist=check_fits_file)


class PixelatedRegularGrid(Grid):
    """Class that represents a grid of values defined on a regular, Cartesian grid.

    Parameters
    ----------
    fits_path : str
        Path to the FITS file in which the values (and perhaps the coordinates)
        are stored. This should be relative to the final COOLEST template file.
    field_of_view_x : Tuple[float], optional
        2-tuple holding the extremal coordinates of the coordinates grid 
        along the x direction (i.e., left side of the leftmost pixel and 
        rightside of the rightmost pixel), by default (0, 0)
    field_of_view_y : Tuple[float], optional
        2-tuple holding the extremal coordinates of the coordinates grid 
        along the y direction (i.e., bottom side of the lower pixel and 
        top side of the upper pixel), by default (0, 0)
    num_pix_x : int, optional
        Number of pixels along the x direction, by default 0
    num_pix_y : int, optional
        Number of pixels along the y direction, by default 0
    **kwargs_file : dic, optional
        Any remaining keyword arguments for FitsFile
    """
    
    def __init__(self, 
                 fits_path: str = None,
                 field_of_view_x: Tuple[float] = (0, 0),
                 field_of_view_y: Tuple[float] = (0, 0),
                 num_pix_x: int = 0, 
                 num_pix_y: int = 0,
                 **kwargs_file) -> None:
        super().__init__(fits_path, **kwargs_file)
        self.set_grid(None, field_of_view_x, field_of_view_y, 
                      num_pix_x, num_pix_y, 
                      check_fits_file=kwargs_file.get('check_fits_file', True))

    @property
    def shape(self):
        return (self.num_pix_x, self.num_pix_y)

    @property
    def pixel_size(self):
        if self.num_pix_x == 0 or self.num_pix_y == 0:
            return 0.
        pix_size_x = np.abs(self.field_of_view_x[0] - self.field_of_view_x[1]) / self.num_pix_x
        pix_size_y = np.abs(self.field_of_view_y[0] - self.field_of_view_y[1]) / self.num_pix_y
        assert pix_size_x == pix_size_y, "Regular grid must have square pixels"
        return pix_size_x

    def set_grid(self, fits_path, 
                 field_of_view_x, field_of_view_y,
                 num_pix_x=0, num_pix_y=0,
                 check_fits_file=True):
        """Set / replace the FitsFile object associated with the Grid.
        This is useful to set the FITS file after the Grid instance
        has been created.

        See class constructor for parameter descriptions.
        """
        super().set_grid(fits_path, check_fits_file)
        self.field_of_view_x = field_of_view_x
        self.field_of_view_y = field_of_view_y
        if self.fits_file.exists() and check_fits_file:
            self.num_pix_x, self.num_pix_y = self.read_fits()
            # if number of pixels is also given, check that it is consistent
            if num_pix_x != 0 and self.num_pix_x != num_pix_x:
                raise ValueError("Given number of pixels in x direction "
                                 "is inconsistent with the fits file")
            if num_pix_y != 0 and self.num_pix_y != num_pix_y:
                raise ValueError("Given number of pixels in y direction "
                                 "is inconsistent with the fits file")
        else:
            self.num_pix_x, self.num_pix_y = num_pix_x, num_pix_y

    def read_fits(self):
        """Read the content of the FITS file and extract the necessary Grid attributes.

        Returns
        -------
        (num_pix_x, num_pix_y)
            Number of pixels along each axis.
        """
        array, header = self.fits_file.read()
        array_shape = array.shape
        if array_shape != (header['NAXIS1'], header['NAXIS2']):
            warnings.warn("Image dimensions do not match the FITS header")
        return array_shape

    def get_pixels(self, directory=None):
        """Get the pixel (z) values of the regular grid from the FITS file.
        If the attribute FITS path is a relative one, it needs the absolute
        directory to read the FITS file.

        Parameters
        ----------
        directory : str, optional
            Absolute directory of the FITS file, by default None

        Returns
        -------
        ndarray
            2D array of pixel values associated to the regular grid.
        """
        array, _ = self.fits_file.read(directory=directory)
        return array


class IrregularGrid(Grid):
    """Class that represents an irregular set of values and their coordinates.

    Parameters
    ----------
    fits_path : str
        Path to the FITS file in which the values (and perhaps the coordinates)
        are stored. This should be relative to the final COOLEST template file.
    field_of_view_x : Tuple[float], optional
        2-tuple holding the minimum and maximum values of the x coordinates vector, 
        by default (0, 0)
    field_of_view_y : Tuple[float], optional
         2-tuple holding the minimum and maximum values of the y coordinates vector, 
        by default (0, 0)
    num_pix : int, optional
        Number of coordinates points, by default 0
    **kwargs_file : dic, optional
        Any remaining keyword arguments for FitsFile
    """

    def __init__(self, 
                 fits_path: str = None,
                 field_of_view_x: Tuple[float] = (0, 0),
                 field_of_view_y: Tuple[float] = (0, 0),
                 num_pix: int = 0,
                 **kwargs_file) -> None:
        super().__init__(fits_path, **kwargs_file)
        self.set_grid(None, 
                      field_of_view_x=field_of_view_x, 
                      field_of_view_y=field_of_view_y, 
                      num_pix=num_pix)

    def set_grid(self, fits_path, 
                 field_of_view_x=(0, 0), field_of_view_y=(0, 0),
                 num_pix=0, check_fits_file=True):
        """Set / replace the FitsFile object associated with the Grid.
        This is useful to set associate the FITS file after the Grid instance
        has been created.

        See class constructor for parameter descriptions.
        """
        super().set_grid(fits_path, check_fits_file)
        if self.fits_file.exists():
            self.field_of_view_x, self.field_of_view_y, self.num_pix = self.read_fits()
            if num_pix != 0 and self.num_pix != num_pix:
                raise ValueError("Given number of pixels is inconsistent with the fits file")
            if field_of_view_x != (0, 0) and field_of_view_x == self.field_of_view_x:
                raise ValueError("Given field of view along x direction is inconsistent with the fits file")
            if field_of_view_y != (0, 0) and field_of_view_y == self.field_of_view_y:
                raise ValueError("Given field of view along y direction is inconsistent with the fits file")
        else:
            self.field_of_view_x = field_of_view_x
            self.field_of_view_y = field_of_view_y
            self.num_pix = num_pix

    def read_fits(self):
        """Read the content of the FITS file and extract the necessary Grid attributes.

        Returns
        -------
        field_of_view_x, field_of_view_y, num_pix
            Field of view and number of pxiels.
        """
        x, y, z = self.get_xyz()
        num_pix = len(z)
        #assert self.num_pix == len(z), "Given number of grid points does not match the number of .fits table rows!"
        # Here we may want to check/report the overlap between the given field of view and the square encompassing the irregular grid
        field_of_view_x = (min(x), max(x))
        field_of_view_y = (min(y), max(y))
        return field_of_view_x, field_of_view_y, num_pix

    def get_xyz(self, directory=None):
        """Get the x, y, z values of the irregular grid from the FITS file.
        If the attribute FITS path is a relative one, it needs the absolute
        directory to read the FITS file.

        Parameters
        ----------
        directory : str, optional
            Absolute directory of the FITS file, by default None

        Returns
        -------
        ndarray, ndarray, ndarray
            x, y and z arrays
        """
        data, _ = self.fits_file.read(directory=directory)
        x = data.field(0)
        y = data.field(1)
        z = data.field(2)
        return x, y, z



__author__ = 'aymgal'


import inspect
import json
from copy import deepcopy


def get_class_names(instance):
    class_names = [c.__name__ for c in inspect.getmro(instance.__class__)]
    return class_names[0], ' < '.join(class_names[:-1])


def filter_dict(dictionary, exclude_keys=None):
    if exclude_keys is None:
        return dictionary
    dictionary_ = deepcopy(dictionary)
    for key in dictionary.keys():
        if key in exclude_keys or key[0] == '_':
            del dictionary_[key]
        elif isinstance(dictionary[key], dict):
            dictionary_[key] = filter_dict(dictionary_[key], exclude_keys)
    return dictionary_


# class Fields(object):
#     """Essentially a wrapper of inspect.signature
#     to store argument names and default value in a readable json format.
#     This is mainly for the documentation website"""
#     def __init__(self, func):
#         signature = inspect.signature(func)
#         for key, parameter in signature.parameters.items():
#             setattr(self, str(parameter.name), str(parameter.default))


class APIBaseObject(object):
    """Abstract class on which all COOLEST template classes are based.
    This handles the filtering of attributes and automatic documentation attributes.

    
    """

    def __init__(self):
        """Creates a `documentation` attribute based on the docstring 
        of the child class.
        """
        #self.type, self._api_inheritance = get_class_names(self)
        if not hasattr(self, 'documentation') or self.documentation is None:
            if self.__doc__ is not None:
                self.documentation = self.__doc__.strip()
            else:
                self.documentation = ""
        # self.fields = Fields(self.__init__)
        
    def to_JSON(self, indent=2, exclude_keys=None):
        """Returns a JSON representation of `self`, filtering out specific attributes
        based on their name.

        Parameters
        ----------
        indent : int, optional
            Indentation to be used in the JSON representation, by default 2
        exclude_keys : list, optional
            List of attribute names to be excluded from the JSON representation
            (see `standard` submodule for examples), by default None
        """
        return json.dumps(self, default=lambda o: filter_dict(o.__dict__, exclude_keys), 
                          sort_keys=True, indent=indent)



__author__ = 'aymgal'

from typing import Tuple

from coolest.template.classes.base import APIBaseObject
from coolest.template.classes.lensing_entity import LensingEntity
from coolest.template.classes.parameter import Parameter
from coolest.template.classes import util


class LensingEntityList(list, APIBaseObject):
    """The list of components that define the lensing system.
    In COOLEST, a `LensingEntity` is an instance Galaxy or a MassField.

    Note that unique identifiers (IDs) for each profile and parameters will be 
    generated at instantiation time.

    - A given profile has a unique IDs with the following pattern:

    `{entity index}-{{massfield} or {galaxy}}-{{mass} or {light}}-{profile index}_{profile name}`

    - A given parameter has the same ID as above, just with the parameter name at the end:

    `{entity index}-{{massfield} or {galaxy}}-{{mass} or {light}}-{profile index}_{profile name}-{parameter name}`

    Parameters
    ----------
    *entities : LensingEntity instances
        As many LensingEntity instances as required
    """

    def __init__(self, *entities: Tuple[LensingEntity]):
        list.__init__(self, entities)
        APIBaseObject.__init__(self)
        self._create_all_ids()

    def get_parameters(self, with_name=None, with_fixed=True):
        """Returns the list of either all parameters in the model, 
        or only a subset of them for parameters with a specific name.

        Parameters
        ----------
        with_name : str, optional
            Parameter for which we want to get all corresponding IDs (default: None).
        with_fixed : bool, optional
            If True, includes also fixed parameters (default: True).

        Returns
        -------
        list
            List of parameter instances
        """
        def _selected(param_name, param):
            # below we check that `param` is a Parameter instance because Grid-like parameters
            # do not have (yet) the possibility to be fixed (no fixed attribute).
            is_fixed = False if not isinstance(param, Parameter) else param.fixed
            ignored_if_fixed = not with_fixed and is_fixed
            if (with_name is None) or (param_name == with_name):
                return False if ignored_if_fixed else True
            else:
                return False
        param_list = []
        for entity in self:
            for model_type in ('light', 'mass'):
                model = getattr(entity, f'{model_type}_model', None)
                if model is not None:
                    for profile in model:
                        for param_name, param in profile.parameters.items():
                            if _selected(param_name, param):
                                param_list.append(param)
        return param_list
    
    def get_parameter_ids(self, with_name=None, with_fixed=True):
        """Returns the list of either all parameter IDs in the model, 
        or only a subset of them for parameters with a specific name.

        Parameters
        ----------
        with_name : str, optional
            Parameter for which we want to get all corresponding IDs (default: None).
        with_fixed : bool, optional
            If True, includes also fixed parameters (default: True).

        Returns
        -------
        list
            List of IDs (strings)
        """
        return [p.id for p in self.get_parameters(with_name=with_name, with_fixed=with_fixed)]
    
    def get_parameter_from_id(self, param_id):
        """Returns the Parameter instance that has the given parameter ID, or None.

        Parameters
        ----------
        param_id : str
            Parameter ID

        Returns
        -------
        coolest.template.classes.parameters.Parameter or None
            Instance of a Parameter with ID equal to `param_ID`
        """
        # NOTE: it can be very inefficient for a large number of lensing entities
        for entity in self:
            for model_type in ('light', 'mass'):
                model = getattr(entity, f'{model_type}_model', None)
                if model is not None:
                    for profile in model:
                        for param in profile.parameters.values():
                            if param.id == param_id:
                                return param
        # if the following line is reached, then no ID has been found
        return None
        #raise ValueError("Parameter with ID '{param_id}' not found in any lensing entity.")

    def _create_all_ids(self):
        for i, entity in enumerate(self):
            for model_type in ('light', 'mass'):
                model = getattr(entity, f'{model_type}_model', None)
                if model is not None:
                    for j, profile in enumerate(model):
                        if entity.type == 'Galaxy':
                            profile_id = util.galaxy_profile_to_id(model_type, profile.type, j, i)
                        elif entity.type == 'MassField':
                            profile_id = util.mass_field_profile_to_id(profile.type, j, i)
                        profile.id = profile_id
                        for param_name, parameter in profile.parameters.items():
                            param_id = util.parameter_to_id(param_name, profile.id)
                            parameter.id = param_id



__author__ = 'aymgal'

from coolest.template.classes.lensing_entity import LensingEntity
from coolest.template.classes.mass_light_model import MassModel, LightModel


class Galaxy(LensingEntity):
    """Class that defines a Galaxy entity (lens galaxy and/or source galaxy).

    Parameters
    ----------
    name : str
        Name associated to the galaxy.
    redshift : float
        Redshift of the galaxy, if any.
    light_model : LightModel, optional
        Light model instance describing the surface brightness of the galaxy, by default None
    mass_model : MassModel, optional
        Mass model instance describing the mass distribution of the galaxy, by default None
    """

    def __init__(self,
                 name: str,
                 redshift: float,
                 light_model: LightModel = None,
                 mass_model: MassModel = None) -> None:
        super().__init__(name, redshift, mass_model=mass_model)
        if light_model is None:
            light_model = LightModel()
        self.light_model = light_model



__author__ = 'aymgal'

from typing import Tuple

from coolest.template.classes.base import APIBaseObject


SUPPORTED_CHOICES = [
    'imaging_data', 
    'time_delays',
    'magnification_ratios',
    'image_plane_position',
    'source_plane_position',
]


class LikelihoodList(list, APIBaseObject):

    """
    A list all likelihood terms used during lens model parameters optimization. 
    Only applicable in the 'MAP' mode.
    """

    def __init__(self, 
                 *likelihood_types: Tuple[str]) -> None:
        for ll_type in likelihood_types:
            if ll_type not in SUPPORTED_CHOICES:
                raise ValueError(f"Likelihood type '{ll_type}' is not supported.")
        list.__init__(self, likelihood_types)
        APIBaseObject.__init__(self)


__author__ = 'aymgal'

from coolest.template.classes.base import APIBaseObject
from coolest.template.classes.mass_light_model import MassModel


__all__ = [
    'LensingEntity',
]

SUPPORTED_CHOICES = ['Galaxy', 'MassField']


class LensingEntity(APIBaseObject):
    """Abstract lensing entity, typically a Galaxy or an MassField object.

    Parameters
    ----------
    name : str
        Name of the entity (e.g., the name of the galaxy)
    redshift : float
        Redshift of the entity, if known / assumed
    mass_model : MassModel, optional
        Mass model (i.e., a set of mass profiles), by default None

    Raises
    ------
    ValueError
        If the provided redshift is negative.
    """

    def __init__(self,
                 name: str,
                 redshift: float,
                 mass_model: MassModel = None) -> None:
        self.type = self.__class__.__name__  # name of the children class
        self.name = name
        if redshift is not None and redshift < 0:
            raise ValueError("Redshift cannot be negative.")
        self.redshift = redshift
        if mass_model is None:
            mass_model = MassModel()
        self.mass_model = mass_model
        super().__init__()



__author__ = 'aymgal'

import os
import numpy as np
import math
import logging
from scipy import signal
import pandas as pd
from functools import partial

from coolest.api import util


# logging settings
logging.getLogger().setLevel(logging.WARNING)


class BaseComposableModel(object):
    """Given a COOLEST object, evaluates a selection of mass or light profiles.
    This class serves as parent for more specific classes and should not be 
    instantiated by the user.

    Parameters
    ----------
    model_type : str
        Either 'light_model' or 'mass_model'
    coolest_object : COOLEST
        COOLEST instance
    coolest_directory : str, optional
        Directory which contains the COOLEST template, by default None
    load_posterior_samples : bool, optional
        If True, and if the COOLEST metadata provides it, the constructor will
        attempt to load the chain file containing posterior samples, in addition
        to point estimates for each profile parameters. Default is False.
    entity_selection : list, optional
        List of indices of the lensing entities to consider; If None, 
        selects the first entity which has a model of type model_type, by default None
    profile_selection : list, optional
        List of either lists of indices, or 'all', for selecting which (mass or light) profile 
        of a given lensing entity to consider. If None, selects all the 
        profiles of within the corresponding entity, by default None

    Raises
    ------
    ValueError
        No valid entity found or no profiles found.
    """
        
    _chain_key = "chain_file_name"
    _supported_eval_modes = ('point', 'posterior')

    def __init__(self, model_type, 
                 coolest_object, coolest_directory=None, 
                 load_posterior_samples=False,
                 entity_selection=None, profile_selection=None):
        if entity_selection is None:
            # finds the first entity that has a 'model_type' profile
            entity_selection = None
            for i, entity in enumerate(coolest_object.lensing_entities):
                if model_type == 'light_model' \
                    and entity.type == 'galaxy' \
                    and len(entity.light_model) > 0:
                    entity_selection = [i]
                    break
                elif model_type == 'mass_model' \
                    and len(entity.mass_model) > 0:
                    entity_selection = [i]
                    break
            if entity_selection is None:
                raise ValueError("No lensing entity with light profiles have been found")
            else:
                logging.warning(f"Found valid profile for lensing entity (index {i}) for model type '{model_type}'")
        if profile_selection is None:
            profile_selection = 'all'
        entities = coolest_object.lensing_entities
        self.directory = coolest_directory
        self._posterior_bool, self._csv_path = False, None
        if load_posterior_samples:
            metadata = coolest_object.meta
            if self._chain_key not in metadata:
                logging.warning(f"Metadata key '{self._chain_key}' is missing "
                                f"from COOLEST template, hence no posterior samples "
                                f"will be loaded.")
            else:
                self._posterior_bool = True
                self._csv_path = os.path.join(self.directory, metadata[self._chain_key])
        self.setup_profiles_and_params(model_type, entities, 
                                        entity_selection, profile_selection)
        self.num_profiles = len(self.profile_list)
        if self.num_profiles == 0:
            raise ValueError("No profile has been selected!")

    def setup_profiles_and_params(self, model_type, entities, 
                                  entity_selection, profile_selection):
        profile_list = []
        param_list, post_param_list = [], []
        info_list = []
        for i, entity in enumerate(entities):
            if self._selected(i, entity_selection):
                if model_type == 'light_model' and entity.type == 'external_shear':
                    raise ValueError(f"External shear (entity index {i}) has no light model")
                for j, profile in enumerate(getattr(entity, model_type)):
                    if self._selected(j, profile_selection):
                        if 'Grid' in profile.type:
                            if self.directory is None:
                                raise ValueError("The directory in which the COOLEST file is located "
                                                 "must be provided for loading FITS files.")
                            params, fixed_params = self._get_grid_params(profile, self.directory)
                            profile_list.append(self._get_api_profile(model_type, profile, *fixed_params))
                            post_params = None  # TODO: support samples for grid parameters
                        else:
                            params, post_params = self._get_regular_params(
                                profile, samples_file_path=self._csv_path
                            )
                            profile_list.append(self._get_api_profile(model_type, profile))
                        param_list.append(params)
                        post_param_list.append(post_params)
                        info_list.append((entity.name, entity.redshift))
        self.profile_list = profile_list
        self.param_list = param_list
        self.info_list = info_list
        if self._posterior_bool is True:
            post_param_list, post_weights = self._finalize_post_samples(post_param_list, self._csv_path)
            self.post_param_list = post_param_list
            self.post_weights = np.array(post_weights)
        else:
            self.post_param_list = None
            self.post_weights = None

    def estimate_center(self):
        # TODO: improve this (for now simply considers the first profile that has a center)
        for profile, params in zip(self.profile_list, self.param_list):
            if 'center_x' in params:
                center_x = params['center_x']
                center_y = params['center_y']
                logging.info(f"Picked center from profile '{profile.type}'")
                return center_x, center_y
        raise ValueError("Could not estimate a center from the composed model")

    @staticmethod
    def _get_api_profile(model_type, profile_in, *extra_profile_args):
        """
        Takes as input a light profile from the template submodule
        and instantites the corresponding profile from the API submodule
        """
        if model_type == 'light_model':
            from coolest.api.profiles import light
            ProfileClass = getattr(light, profile_in.type)
        elif model_type == 'mass_model':
            from coolest.api.profiles import mass
            ProfileClass = getattr(mass, profile_in.type)
        return ProfileClass(*extra_profile_args)

    @staticmethod
    def _get_regular_params(profile_in, samples_file_path=None):
        parameters = {}  # best-fit values
        samples = {} if samples_file_path else None  # posterior samples
        for name, param in profile_in.parameters.items():
            parameters[name] = param.point_estimate.value
            if samples is not None:
                # read just the column corresponding to the parameter ID
                column = pd.read_csv(
                    samples_file_path, 
                    usecols=[param.id], 
                    delimiter=',',
                )
                # TODO: take into account probability weights from nested sampling runs!
                samples[name] = list(column[param.id])
        return parameters, samples

    @staticmethod
    def _get_grid_params(profile_in, fits_dir):
        param_in = profile_in.parameters['pixels']
        if profile_in.type == 'PixelatedRegularGrid':
            data = param_in.get_pixels(directory=fits_dir)
            parameters = {'pixels': data}
            fov_x = param_in.field_of_view_x
            fov_y = param_in.field_of_view_y
            npix_x = param_in.num_pix_x
            npix_y = param_in.num_pix_y
            fixed_parameters = (fov_x, fov_y, npix_x, npix_y)

        elif profile_in.type == 'IrregularGrid':
            x, y, z = param_in.get_xyz(directory=fits_dir)
            parameters = {'x': x, 'y': y, 'z': z}
            fov_x = param_in.field_of_view_x
            fov_y = param_in.field_of_view_y
            npix = param_in.num_pix
            fixed_parameters = (fov_x, fov_y, npix)
        return parameters, fixed_parameters
    
    @staticmethod
    def _finalize_post_samples(param_list_of_samples, samples_file_path):
        """
        Takes as input the samples grouped at the leaves of the nested container structure,
        and returns a list of items each organized as self.param_list
        """
        num_profiles = len(param_list_of_samples)
        profile_0 = param_list_of_samples[0]
        num_samples = len(profile_0[list(profile_0.keys())[0]])
        samples_of_param_list = [
            [{} for _ in range(num_profiles)] for _ in range(num_samples)
        ]
        for i in range(num_samples):
            for k in range(num_profiles):
                for key in param_list_of_samples[k].keys():
                    samples_of_param_list[i][k][key] = param_list_of_samples[k][key][i]
        # also load and return the probability weights
        # read just the column corresponding to the parameter ID
        weights_key = 'probability_weights'
        column = pd.read_csv(
            samples_file_path, 
            usecols=[weights_key], 
            delimiter=',',
        )
        weights_list = list(column[weights_key])
        return samples_of_param_list, weights_list

    @staticmethod
    def _selected(index, selection):
        if isinstance(selection, str) and selection.lower() == 'all':
            return True
        elif isinstance(selection, (list, tuple, np.ndarray)) and index in selection:
            return True
        elif isinstance(selection, (int, float)) and int(selection) == index:
            return True
        return False

    def _check_eval_mode(self, mode):
        if mode not in self._supported_eval_modes:
            raise NotImplementedError(
                f"Only evaluation modes "
                f"{self._supported_eval_modes} are supported "
                f"(received '{mode}')."
        )


class ComposableLightModel(BaseComposableModel):
    """Given a COOLEST object, evaluates a selection of entity and their light profiles.

    Parameters
    ----------
    coolest_object : COOLEST
        COOLEST instance
    coolest_directory : str, optional
        Directory which contains the COOLEST template, by default None
    entity_selection : list, optional
        List of indices of the lensing entities to consider; If None, 
        selects the first entity that has a light model, by default None
    profile_selection : list, optional
        List of either lists of indices, or 'all', for selecting which light profile 
        of a given lensing entity to consider. If None, selects all the 
        profiles of within the corresponding entity, by default None

    Raises
    ------
    ValueError
        No valid entity found or no profiles found.
    """

    def __init__(self, coolest_object, coolest_directory=None, **kwargs_selection):
        super().__init__('light_model', coolest_object, 
                         coolest_directory=coolest_directory,
                         **kwargs_selection)
        pixel_size = coolest_object.instrument.pixel_size
        if pixel_size is None:
            self.pixel_area = 1.
        else:
            self.pixel_area = pixel_size**2

    def surface_brightness(self, return_extra=False):
        """Returns the surface brightness as stored in the COOLEST file"""
        if self.num_profiles > 1:
            logging.warning("When more than a single light profile has been selected, "
                            "the method `surface_brightness()` only considers the first profile")
        profile = self.profile_list[0]
        values = profile.surface_brightness(**self.param_list[0])
        if return_extra:
            extent = profile.get_extent()
            coordinates = profile.get_coordinates()
            return values, extent, coordinates
        return values

    def evaluate_surface_brightness(self, x, y):
        """Evaluates the surface brightness at given coordinates"""
        image = np.zeros_like(x)
        for k, (profile, params) in enumerate(zip(self.profile_list, self.param_list)):
            flux_k = profile.evaluate_surface_brightness(x, y, **params)
            if profile.units == 'per_ang':
                flux_k *= self.pixel_area
            image += flux_k
        return image


class ComposableMassModel(BaseComposableModel):
    """Given a COOLEST object, evaluates a selection of entity and their mass profiles.

    Parameters
    ----------
    coolest_object : COOLEST
        COOLEST instance
    coolest_directory : str, optional
        Directory which contains the COOLEST template, by default None
    entity_selection : list, optional
        List of indices of the lensing entities to consider; If None, 
        selects the first entity that has a mass model, by default None
    profile_selection : list, optional
        List of either lists of indices, or 'all', for selecting which mass profile 
        of a given lensing entity to consider. If None, selects all the 
        profiles of within the corresponding entity, by default None

    Raises
    ------
    ValueError
        No valid entity found or no profiles found.
    """

    def __init__(self, coolest_object, coolest_directory=None, 
                 load_posterior_samples=False,
                 **kwargs_selection):
        super().__init__('mass_model', coolest_object, 
                         coolest_directory=coolest_directory,
                         load_posterior_samples=load_posterior_samples,
                         **kwargs_selection)

    def evaluate_potential(self, x, y, mode='point', last_n_samples=None):
        """Evaluates the lensing potential field at given coordinates"""
        self._check_eval_mode(mode)
        if mode == 'point' or self._posterior_bool is False:
            return self._eval_pot_point(x, y, self.param_list)
        elif mode == 'posterior':
            return self._eval_pot_posterior(x, y, self.post_param_list, last_n_samples)

    def _eval_pot_point(self, x, y, param_list):
        psi = np.zeros_like(x)
        for k, profile in enumerate(self.profile_list):
            psi += profile.potential(x, y, **param_list[k])
        return psi
    
    def _eval_pot_posterior(self, x, y, param_list, last_n_samples):
        # map the point function at each sample
        use_all_samples = last_n_samples is None or last_n_samples <= 0
        val_list = param_list if use_all_samples else param_list[-last_n_samples:]
        mapped = map(partial(self._eval_pot_point, x, y), val_list)
        return np.array(list(mapped))
    
    def fermat_potential(self, x, y, x_src, y_src, mode='point', last_n_samples=None):
        """Computes the Fermat potential for image (x, y) and source position (x_src, y_src)
        """
        # gravitational term
        psi = self.evaluate_potential(x, y, mode=mode, last_n_samples=last_n_samples)
        # geometric term
        geo = ((x - x_src)**2 + (y - y_src)**2) / 2.
        geo = np.broadcast_to(geo, psi.shape)  # makes sure geo has same shape as psi
        return geo - psi
    
    def evaluate_deflection(self, x, y):
        """Evaluates the lensing deflection field at given coordinates"""
        alpha_x, alpha_y = np.zeros_like(x), np.zeros_like(x)
        for k, (profile, params) in enumerate(zip(self.profile_list, self.param_list)):
            a_x, a_y = profile.deflection(x, y, **params)
            alpha_x += a_x
            alpha_y += a_y
        return alpha_x, alpha_y

    def evaluate_convergence(self, x, y):
        """Evaluates the lensing convergence (i.e., 2D mass density) at given coordinates"""
        kappa = np.zeros_like(x)
        for k, (profile, params) in enumerate(zip(self.profile_list, self.param_list)):
            kappa += profile.convergence(x, y, **params)
        return kappa

    def evaluate_hessian(self, x, y):
        """Evaluates the lensing Hessian components at given coordinates"""
        H_xx_sum = np.zeros_like(x)
        H_xy_sum = np.zeros_like(x)
        H_yx_sum = np.zeros_like(x)
        H_yy_sum = np.zeros_like(x)
        for k, (profile, params) in enumerate(zip(self.profile_list, self.param_list)):
            H_xx, H_xy, H_yx, H_yy = profile.hessian(x, y, **params)
            H_xx_sum += H_xx
            H_xy_sum += H_xy
            H_yx_sum += H_yx
            H_yy_sum += H_yy
        return H_xx_sum, H_xy_sum, H_yx_sum, H_yy_sum
    
    def evaluate_jacobian(self, x, y):
        """Evaluates the lensing Jacobian (d beta / d theta)  at given coordinates"""
        H_xx, H_xy, H_yx, H_yy = self.evaluate_hessian(x, y)
        A = np.array([[1 - H_xx, -H_xy], [-H_yx, 1 - H_yy]])
        return A
    
    def evaluate_magnification(self, x, y):
        """Evaluates the lensing magnification at given coordinates"""
        H_xx, H_xy, H_yx, H_yy = self.evaluate_hessian(x, y)
        det_A = (1 - H_xx) * (1 - H_yy) - H_xy*H_yx
        mu = 1. / det_A
        return mu

    def ray_shooting(self, x, y):
        """evaluates the lens equation beta = theta - alpha(theta)"""
        alpha_x, alpha_y = self.evaluate_deflection(x, y)
        x_rs, y_rs = x - alpha_x, y - alpha_y
        return x_rs, y_rs


class ComposableLensModel(object):
    """Given a COOLEST object, evaluates a selection of entity and 
    their mass and light profiles, typically to construct an image of the lens.

    Parameters
    ----------
    coolest_object : COOLEST
        COOLEST instance
    coolest_directory : str, optional
        Directory which contains the COOLEST template, by default None
    entity_selection : list, optional
        List of indices of the lensing entities to consider; If None, 
        selects the first entity that has a light/mass model, by default None
    profile_selection : list, optional
        List of either lists of indices, or 'all', for selecting which light/mass profile 
        of a given lensing entity to consider. If None, selects all the 
        profiles of within the corresponding entity, by default None

    Raises
    ------
    ValueError
        No valid entity found or no profiles found.
    """

    def __init__(self, coolest_object, coolest_directory=None, 
                 kwargs_selection_source=None, kwargs_selection_lens_mass=None):
        self.coolest = coolest_object
        self.coord_obs = util.get_coordinates(self.coolest)
        self.directory = coolest_directory
        if kwargs_selection_source is None:
            kwargs_selection_source = {}
        if kwargs_selection_lens_mass is None:
            kwargs_selection_lens_mass = {}
        self.lens_mass = ComposableMassModel(coolest_object, 
                                             coolest_directory,
                                             **kwargs_selection_lens_mass)
        self.source = ComposableLightModel(coolest_object, 
                                          coolest_directory,
                                          **kwargs_selection_source)

    def model_image(self, supersampling=5, convolved=True, super_convolution=True):
        """generates an image of the lens based on the selected model components"""
        obs = self.coolest.observation
        psf = self.coolest.instrument.psf
        if convolved is True and psf.type == 'PixelatedPSF':
            scale_factor = obs.pixels.pixel_size / psf.pixels.pixel_size
            supersampling_conv = int(round(scale_factor))
            if not math.isclose(scale_factor, supersampling_conv):
                raise ValueError(f"PSF supersampling ({scale_factor}) not close to an integer?")
            if supersampling_conv < 1:
                raise ValueError("PSF pixel size smaller than data pixel size")
        if supersampling < 1:
            raise ValueError("Supersampling must be >= 1")
        if convolved is True and supersampling_conv > supersampling:
            supersampling = supersampling_conv
            logging.warning(f"Supersampling adapted to the PSF pixel size ({supersampling})")
        coord_eval = self.coord_obs.create_new_coordinates(pixel_scale_factor=1./supersampling)
        x, y = coord_eval.pixel_coordinates
        image = self.evaluate_lensed_surface_brightness(x, y)
        if convolved is True:
            if psf.type != 'PixelatedPSF':
                raise NotImplementedError
            kernel = psf.pixels.get_pixels(directory=self.directory)
            kernel_sum = kernel.sum()
            if not math.isclose(kernel_sum, 1., abs_tol=1e-3):
                kernel /= kernel_sum
                logging.warning(f"PSF kernel is not normalized (sum={kernel_sum}), "
                                f"so it has been normalized before convolution")
            if np.isnan(image).any():
                np.nan_to_num(image, copy=False, nan=0., posinf=None, neginf=None)
                logging.warning("Found NaN values in image prior to convolution; "
                                "they have been replaced by zeros.")
            if super_convolution and supersampling_conv == supersampling:
                # first convolve then downscale
                image = signal.fftconvolve(image, kernel, mode='same')
                image = util.downsampling(image, factor=supersampling)
            else:
                # first downscale then convolve
                image = util.downsampling(image, factor=supersampling)
                image = signal.fftconvolve(image, kernel, mode='same')
        elif supersampling > 1:
            image = util.downsampling(image, factor=supersampling)
        return image, self.coord_obs

    def model_residuals(self, mask=None, **model_image_kwargs):
        """computes the normalized residuals map as (data - model) / sigma"""
        model, _ = self.model_image(**model_image_kwargs)
        data = self.coolest.observation.pixels.get_pixels(directory=self.directory)
        noise = self.coolest.observation.noise
        if noise.type != 'NoiseMap':
            raise NotImplementedError
        sigma = noise.noise_map.get_pixels(directory=self.directory)
        if mask is None:
            mask = np.ones_like(model)
        return ((data - model) / sigma) * mask, self.coord_obs

    def evaluate_lensed_surface_brightness(self, x, y):
        """Evaluates the surface brightness of a lensed source at given coordinates"""
        # ray-shooting
        x_rs, y_rs = self.ray_shooting(x, y)
        # evaluates at ray-shooted coordinates
        lensed_image = self.source.evaluate_surface_brightness(x_rs, y_rs)
        return lensed_image

    def ray_shooting(self, x, y):
        """evaluates the lens equation beta = theta - alpha(theta)"""
        return self.lens_mass.ray_shooting(x, y)



__author__ = 'aymgal'

from coolest.template.classes.profile import Profile, AnalyticalProfile
from coolest.template.classes.parameter import (NonLinearParameter, 
                                                LinearParameter, 
                                                LinearParameterSet, 
                                                NonLinearParameterSet,
                                                PixelatedRegularGridParameter,
                                                IrregularGridParameter)
from coolest.template.classes.parameter import DefinitionRange


__all__ = [
    'Sersic',
    'Chameleon',
    'Uniform',
    'Shapelets',
    'LensedPS',
    'PixelatedRegularGrid',
    'IrregularGrid',
]
SUPPORTED_CHOICES = __all__


class Sersic(AnalyticalProfile):
    """Surface brightness described by an elliptical Sérsic profile.

    This profile is described by the following parameters:
    
    - 'I_eff': amplitude at the effective radius
    - 'theta_eff': effective radius, a.k.a. Sersic radius, equal to the half-light radius
    - 'n': Sersic index
    - 'q': axis ratio (semi-major axis / semi-minor axis)
    - 'phi': position angle
    - 'center_x': position along the x coordinate
    - 'center_y': position along the y coordinate
    """
    
    def __init__(self):
        parameters = {
            'I_eff': LinearParameter("Amplitude at the Sersic radius",
                            DefinitionRange(min_value=0.0),
                            latex_str=r"$I_{\rm Sersic}$"),
            'theta_eff': NonLinearParameter("Sersic (half-light) radius",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$R_{\rm Sersic}$"),
            'n': NonLinearParameter("Sersic index",
                               DefinitionRange(min_value=0.5, max_value=10.0),
                               latex_str=r"$n_{\rm Sersic}$"),
            'q': NonLinearParameter("Axis ratio, semi-minor axis / semi-major axis",
                                     DefinitionRange(min_value=0.0, max_value=1.0),
                                     latex_str=r"$q$"),
            'phi': NonLinearParameter("Position angle",
                                     DefinitionRange(min_value=-90., max_value=90.),
                                     latex_str=r"$\phi$"),
            'center_x': NonLinearParameter("Profile center along x coordinates",
                               DefinitionRange(),
                               latex_str=r"$x_0$"),
            'center_y': NonLinearParameter("Profile center along y coordinates",
                               DefinitionRange(),
                               latex_str=r"$y_0$")
        }
        super().__init__(parameters)


class Chameleon(AnalyticalProfile):
    """Surface brightness described by an elliptical Chameleon profile,
    defined as the difference between two NIE profiles.

    This profile is described by the following parameters:
    
    - 'A': normalization of the profile
    - 's_c': core radius of the innver NIE
    - 's_t': core radius of the outer NIE
    - 'q': axis ratio (semi-major axis / semi-minor axis)
    - 'phi': position angle
    - 'center_x': position along the x coordinate
    - 'center_y': position along the y coordinate
    """
    
    def __init__(self):
        parameters = {
            'A': LinearParameter("Normalization of the profile",
                            DefinitionRange(min_value=0.0),
                            latex_str=r"$A$"),
            'w_c': NonLinearParameter("Core radius of inner NIE",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$w{\rm c}$"),
            'w_t': NonLinearParameter("Core radius of outer NIE",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$w{\rm t}$"),
            'q': NonLinearParameter("Axis ratio, semi-minor axis / semi-major axis",
                                     DefinitionRange(min_value=0.0, max_value=1.0),
                                     latex_str=r"$q$"),
            'phi': NonLinearParameter("Position angle",
                                     DefinitionRange(min_value=-90., max_value=90.),
                                     latex_str=r"$\phi$"),
            'center_x': NonLinearParameter("Profile center along x coordinates",
                               DefinitionRange(),
                               latex_str=r"$x_0$"),
            'center_y': NonLinearParameter("Profile center along y coordinates",
                               DefinitionRange(),
                               latex_str=r"$y_0$")
        }
        super().__init__(parameters)


class Shapelets(AnalyticalProfile):
    """Surface brightness described by a shapelet basis set.
    """
    
    def __init__(self):
        documentation = "Set of shapelet functions"
        parameters = {
            'n_max': NonLinearParameter("Maximum order of the Shapelet decomposition",
                               DefinitionRange(min_value=0),
                               latex_str=r"$\n_{\rm max}$", fixed=True),
            'beta': NonLinearParameter("Shapelet characteristic scale",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$\beta$"),
            'center_x': NonLinearParameter("Shapelets center along x coordinates",
                               DefinitionRange(),
                               latex_str=r"$x_0$"),
            'center_y': NonLinearParameter("Shapelets center along y coordinates",
                               DefinitionRange(),
                               latex_str=r"$y_0$"),
            'amps': LinearParameterSet("Set of amplitude values for each shapelet function",
                               DefinitionRange(),
                               latex_str=r"$A$"),
        }
        super().__init__(parameters)

        
class LensedPS(AnalyticalProfile):
    """Surface brightness of a set of point sources after being lensed. 

    This profile is described by the following parameters:
    
    - 'ra_list': list of coordinates along the x axis
    - 'dec_list': list of coordinates along the y axis
    - 'amps': list of amplitudes
    """

    def __init__(self):
        documentation = "Set of lensed point sources"
        parameters = {
            'ra_list': NonLinearParameterSet("RA positions of the lensed point sources",
                               DefinitionRange(),
                               latex_str=r"$ra$"),
            'dec_list': NonLinearParameterSet("DEC positions of the lensed point sources",
                               DefinitionRange(),
                               latex_str=r"$dec$"),
            'amps': LinearParameterSet("Set of amplitude values for the lensed point sources",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$A$"),
        }
        super().__init__(parameters)


class Uniform(AnalyticalProfile):
    """Uniform surface brightness profile.

    This profile is described by the following parameters:
    
    - 'A': amplitude
    """
    
    def __init__(self):
        documentation = "Uniform sheet of light"
        parameters = {
            'A': LinearParameter("Central amplitude",
                            DefinitionRange(min_value=0.0),
                            latex_str=r"$A$"),
        }
        super().__init__(parameters)


class PixelatedRegularGrid(Profile):
    """Surface brightness described on a regular grid of pixels.

    This profile is described by the following parameters:
    
    - 'pixels': 2D array of pixel values
    """

    def __init__(self):
        parameters = {
            'pixels': PixelatedRegularGridParameter("Pixel values")
        }
        super().__init__(parameters)


class IrregularGrid(Profile):
    """Surface brightness described on an irregular grid of pixels.

    This profile is described by the following parameters:
    
    - 'pixels': set of 2D coordinates and associated pixel values (x, y, z)
    """

    def __init__(self):
        parameters = {
            'pixels': IrregularGridParameter("Pixel positions and values")
        }
        super().__init__(parameters)



"""

This file initializes the coolest module
and provides some basic information about the package.

"""

# Set the package release version
version_info = (0, 1, 7)
__version__ = '.'.join(str(c) for c in version_info)

# Set the package details
__author__ = 'Aymeric Galan & COOLEST contributors'
__email__ = 'aymeric.galan@gmail.com'
__year__ = '2021-2024'
__credits__ = 'COOLEST contributors'
__url__ = 'https://github.com/aymgal/COOLEST'
__description__ = 'Standard for Strong Gravitational Lensing Analyses'
__license__ = 'GPL-3.0'



__author__ = 'aymgal'

# Utility functions to use throughout the API

def galaxy_profile_to_id(model_type, profile_name, profile_idx, entity_idx):
    return f'{entity_idx}-galaxy-{model_type}-{profile_idx}-{profile_name}'

def mass_field_profile_to_id(profile_name, profile_idx, entity_idx):
    return f'{entity_idx}-massfield-mass-{profile_idx}-{profile_name}'

def parameter_to_id(param_name, parent_profile_id):
    return f'{parent_profile_id}-{param_name}'

def unpack_profile_id(id_str):
    entity_idx, _, model_type, profile_idx, profile_name = id_str.split('-')
    return int(entity_idx), str(model_type), int(profile_idx), str(profile_name)

def unpack_parameter_id(id_str):
    entity_idx, _, model_type, profile_idx, profile_name, param_name = id_str.split('-')
    return int(entity_idx), str(model_type), int(profile_idx), str(profile_name), str(param_name)



__author__ = 'aymgal', 'lynevdv', 'gvernard'


import os
import copy
import logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize, LogNorm, TwoSlopeNorm
from matplotlib.colors import ListedColormap
from getdist import plots, chains, MCSamples

from coolest.api.analysis import Analysis
from coolest.api.composable_models import *
from coolest.api import util
from coolest.api import plot_util as plut

import pandas as pd


# matplotlib global settings
plt.rc('image', interpolation='none', origin='lower') # imshow settings

# logging settings
logging.getLogger().setLevel(logging.INFO)

# TODO: separate ParametersPlotter from ModelPlotter to avoid dependencies on getdist


class ModelPlotter(object):
    """Create pyplot panels from a lens model stored in the COOLEST format.

    Parameters
    ----------
    coolest_object : COOLEST
        COOLEST instance
    coolest_directory : str, optional
        Directory which contains the COOLEST template, by default None
    color_bad_values : str, optional
        Color assigned to NaN values (typically negative values in log-scale), 
        by default '#111111' (dark gray)
    """

    def __init__(self, coolest_object, coolest_directory=None, 
                 color_bad_values='#222222'):
        self.coolest = coolest_object
        self._directory = coolest_directory

        self.cmap_flux = copy.copy(plt.get_cmap('magma'))
        self.cmap_flux.set_bad(color_bad_values)

        self.cmap_mag = plt.get_cmap('viridis')
        self.cmap_conv = plt.get_cmap('cividis')
        self.cmap_res = plt.get_cmap('RdBu_r')

        #cmap_colors = self.cmap_flux(np.linspace(0, 1, 256))
        #cmap_colors[0,:] = [0.15, 0.15, 0.15, 1.0]  # Set the color of the very first value to gray
        #self.cmap_flux_mod = ListedColormap(cmap_colors)

    def plot_data_image(self, ax, norm=None, cmap=None, xylim=None,
                        neg_values_as_bad=False, add_colorbar=True):
        """plt.imshow panel with the data image"""
        if cmap is None:
            cmap = self.cmap_flux
        coordinates = util.get_coordinates(self.coolest)
        extent = coordinates.plt_extent
        image = self.coolest.observation.pixels.get_pixels(directory=self._directory)
        ax, im = plut.plot_regular_grid(ax, image, extent=extent, 
                                cmap=cmap, norm=norm,
                                neg_values_as_bad=neg_values_as_bad, 
                                xylim=xylim)
        if add_colorbar:
            cb = plut.nice_colorbar(im, ax=ax, max_nbins=4)
            cb.set_label("flux")
        return image

    def plot_surface_brightness(self, ax, coordinates=None,
                                extent_irreg=None, norm=None, cmap=None, 
                                xylim=None, neg_values_as_bad=True,
                                plot_points_irreg=False, add_colorbar=True,
                                kwargs_light=None,
                                plot_caustics=None, caustics_color='white', caustics_alpha=0.5,
                                coordinates_lens=None, kwargs_lens_mass=None):
        """plt.imshow panel showing the surface brightness of the (unlensed)
        lensing entity selected via kwargs_light (see ComposableLightModel docstring)"""
        if extent_irreg is not None:
            raise ValueError("`extent_irreg` is deprecated; use `xylim` instead.")
        if kwargs_light is None:
            kwargs_light = {}
        light_model = ComposableLightModel(self.coolest, self._directory, **kwargs_light)
        if plot_caustics:
            if kwargs_lens_mass is None:
                raise ValueError("`kwargs_lens_mass` must be provided to compute caustics")
            if coordinates_lens is None:
                coordinates_lens = util.get_coordinates(self.coolest).create_new_coordinates(pixel_scale_factor=0.1)
            # NOTE: here we assume that `kwargs_light` is for the source!
            mass_model = ComposableMassModel(self.coolest, self._directory, **kwargs_lens_mass)
            _, caustics = util.find_all_lens_lines(coordinates_lens, mass_model)
        if cmap is None:
            cmap = self.cmap_flux
        if coordinates is not None:
            x, y = coordinates.pixel_coordinates
            image = light_model.evaluate_surface_brightness(x, y)
            extent = coordinates.plt_extent
            ax, im = plut.plot_regular_grid(ax, image, extent=extent, cmap=cmap,
                                             neg_values_as_bad=neg_values_as_bad, 
                                             norm=norm, xylim=xylim)
        else:
            values, extent_model, coordinates = light_model.surface_brightness(return_extra=True)
            if isinstance(values, np.ndarray) and len(values.shape) == 2:
                image = values
                ax, im = plut.plot_regular_grid(ax, image, extent=extent_model, 
                                        cmap=cmap, 
                                        neg_values_as_bad=neg_values_as_bad,
                                        norm=norm, xylim=xylim)
            else:
                points = values
                if xylim is None:
                    xylim = extent_model
                ax, im = plut.plot_irregular_grid(ax, points, xylim, norm=norm, cmap=cmap, 
                                                   neg_values_as_bad=neg_values_as_bad,
                                                   plot_points=plot_points_irreg)
                image = None
        if plot_caustics:
            for caustic in caustics:
                ax.plot(caustic[0], caustic[1], lw=1, color=caustics_color, alpha=caustics_alpha)
        if add_colorbar:
            cb = plut.nice_colorbar(im, ax=ax, max_nbins=4)
            cb.set_label("flux")
        return image, coordinates

    def plot_model_image(self, ax,
                         norm=None, cmap=None, xylim=None, neg_values_as_bad=False,
                         kwargs_source=None, add_colorbar=True, kwargs_lens_mass=None,
                         **model_image_kwargs):
        """plt.imshow panel showing the surface brightness of the (lensed)
        selected lensing entities (see ComposableLensModel docstring)
        """
        if cmap is None:
            cmap = self.cmap_flux
        lens_model = ComposableLensModel(self.coolest, self._directory,
                                         kwargs_selection_source=kwargs_source,
                                         kwargs_selection_lens_mass=kwargs_lens_mass)
        image, coordinates = lens_model.model_image(**model_image_kwargs)
        extent = coordinates.plt_extent
        ax, im = plut.plot_regular_grid(ax, image, extent=extent, 
                                cmap=cmap,
                                neg_values_as_bad=neg_values_as_bad, 
                                norm=norm, xylim=xylim)
        if add_colorbar:
            cb = plut.nice_colorbar(im, ax=ax, max_nbins=4)
            cb.set_label("flux")
        return image

    def plot_model_residuals(self, ax, mask=None,
                             norm=None, cmap=None, xylim=None, add_chi2_label=False, chi2_fontsize=12,
                             kwargs_source=None, add_colorbar=True, kwargs_lens_mass=None,
                             **model_image_kwargs):
        """plt.imshow panel showing the normalized model residuals image"""
        if cmap is None:
            cmap = self.cmap_res
        if norm is None:
            norm = Normalize(-6, 6)
        lens_model = ComposableLensModel(self.coolest, self._directory,
                                         kwargs_selection_source=kwargs_source,
                                         kwargs_selection_lens_mass=kwargs_lens_mass)
        image, coordinates = lens_model.model_residuals(mask=mask, **model_image_kwargs)
        extent = coordinates.plt_extent
        ax, im = plut.plot_regular_grid(ax, image, extent=extent, 
                                cmap=cmap,
                                neg_values_as_bad=False, 
                                norm=norm, xylim=xylim)
        if add_colorbar:
            cb = plut.nice_colorbar(im, ax=ax, max_nbins=4)
            cb.set_label("(data $-$ model) / noise")
        if add_chi2_label is True:
            num_constraints = np.size(image) if mask is None else np.sum(mask)
            red_chi2 = np.sum(image**2) / num_constraints
            ax.text(0.05, 0.05, r'$\chi^2_\nu$='+f'{red_chi2:.2f}', color='black', alpha=1, 
                    fontsize=chi2_fontsize, va='bottom', ha='left', transform=ax.transAxes,
                    bbox={'color': 'white', 'alpha': 0.6})
        return image

    def plot_convergence(self, ax, coordinates=None,
                         norm=None, cmap=None, xylim=None, neg_values_as_bad=False,
                         add_colorbar=True, kwargs_lens_mass=None):
        """plt.imshow panel showing the 2D convergence map associated to the
        selected lensing entities (see ComposableMassModel docstring)
        """
        if kwargs_lens_mass is None:
            kwargs_lens_mass = {}
        mass_model = ComposableMassModel(self.coolest, self._directory,
                                         **kwargs_lens_mass)
        if cmap is None:
            cmap = self.cmap_conv
        if coordinates is None:
            coordinates = util.get_coordinates(self.coolest)
        extent = coordinates.plt_extent
        x, y = coordinates.pixel_coordinates
        image = mass_model.evaluate_convergence(x, y)
        ax, im = plut.plot_regular_grid(ax, image, extent=extent, 
                                cmap=cmap,
                                neg_values_as_bad=neg_values_as_bad, 
                                norm=norm, xylim=xylim)
        if add_colorbar:
            cb = plut.nice_colorbar(im, ax=ax, max_nbins=4)
            cb.set_label(r"$\kappa$")
        return image
    
    def plot_convergence_diff(
            self, ax, reference_map, relative_error=True,    
            norm=None, cmap=None, xylim=None, coordinates=None,
            add_colorbar=True, kwargs_lens_mass=None,
            plot_crit_lines=False, crit_lines_color='black', crit_lines_alpha=0.5):
        """plt.imshow panel showing the 2D convergence map associated to the
        selected lensing entities (see ComposableMassModel docstring)
        """
        if kwargs_lens_mass is None:
            kwargs_lens_mass = {}
        mass_model = ComposableMassModel(self.coolest, self._directory,
                                         **kwargs_lens_mass)
        if cmap is None:
            cmap = self.cmap_res
        if norm is None:
            norm = Normalize(-1, 1)
        if coordinates is None:
            coordinates = util.get_coordinates(self.coolest)
        if plot_crit_lines:
            critical_lines, _ = util.find_all_lens_lines(coordinates, mass_model)
        extent = coordinates.plt_extent
        x, y = coordinates.pixel_coordinates
        image = mass_model.evaluate_convergence(x, y)
        if relative_error is True:
            diff = (reference_map - image) / reference_map
        else:
            diff = reference_map - image
        ax, im = plut.plot_regular_grid(ax, diff, extent=extent, 
                                cmap=cmap, 
                                norm=norm, xylim=xylim)
        if plot_crit_lines:
            for cline in critical_lines:
                ax.plot(cline[0], cline[1], lw=1, color=crit_lines_color, alpha=crit_lines_alpha)
        if add_colorbar:
            cb = plut.nice_colorbar(im, ax=ax, max_nbins=4)
            cb.set_label(r"$\kappa$")
        return image

    def plot_magnification(self, ax, 
                          norm=None, cmap=None, xylim=None,
                          add_colorbar=True, coordinates=None, kwargs_lens_mass=None):
        """plt.imshow panel showing the 2D magnification map associated to the
        selected lensing entities (see ComposableMassModel docstring)
        """
        if kwargs_lens_mass is None:
            kwargs_lens_mass = {}
        mass_model = ComposableMassModel(self.coolest, self._directory,
                                         **kwargs_lens_mass)
        if cmap is None:
            cmap = self.cmap_mag
        if norm is None:
            norm = Normalize(-10, 10)
        if coordinates is None:
            coordinates = util.get_coordinates(self.coolest)
        x, y = coordinates.pixel_coordinates
        extent = coordinates.plt_extent
        image = mass_model.evaluate_magnification(x, y)
        ax, im = plut.plot_regular_grid(ax, image, extent=extent, 
                                cmap=cmap, 
                                norm=norm, xylim=xylim)
        if add_colorbar:
            cb = plut.nice_colorbar(im, ax=ax, max_nbins=4)
            cb.set_label(r"$\mu$")
        return image

    def plot_magnification_diff(
            self, ax, reference_map, relative_error=True,
            norm=None, cmap=None, xylim=None,
            add_colorbar=True, coordinates=None, kwargs_lens_mass=None):
        """plt.imshow panel showing the (absolute or relative) 
        difference between 2D magnification maps
        """
        if kwargs_lens_mass is None:
            kwargs_lens_mass = {}
        mass_model = ComposableMassModel(self.coolest, self._directory,
                                        **kwargs_lens_mass)
        if cmap is None:
            cmap = self.cmap_res
        if norm is None:
            norm = Normalize(-1, 1)
        if coordinates is None:
            coordinates = util.get_coordinates(self.coolest)
        x, y = coordinates.pixel_coordinates
        extent = coordinates.plt_extent
        image = mass_model.evaluate_magnification(x, y)
        if relative_error is True:
            diff = (reference_map - image) / reference_map
        else:
            diff = reference_map - image
        ax, im = plut.plot_regular_grid(ax, diff, extent=extent, 
                                cmap=cmap,
                                norm=norm, xylim=xylim)
        if add_colorbar:
            cb = plut.nice_colorbar(im, ax=ax, max_nbins=4)
            cb.set_label(r"$\mu$")
        return image


class MultiModelPlotter(object):
    """Wrapper around a set of ModelPlotter instances to produce panels that
    consistently compare different models, evaluated on the same
    coordinates systems.

    Parameters
    ----------
    coolest_objects : list
        List of COOLEST instances
    coolest_directories : list, optional
        List of directories corresponding to each COOLEST instance, by default None
    kwargs_plotter : dict, optional
        Additional keyword arguments passed to ModelPlotter
    """

    def __init__(self, coolest_objects, coolest_directories=None, **kwargs_plotter):
        self.num_models = len(coolest_objects)
        if coolest_directories is None:
            coolest_directories = self.num_models * [None]
        self.plotter_list = []
        for coolest, c_dir in zip(coolest_objects, coolest_directories):
            self.plotter_list.append(ModelPlotter(coolest, coolest_directory=c_dir,
                                                  **kwargs_plotter))

    def plot_surface_brightness(self, axes, **kwargs):
        return self._plot_light_multi('plot_surface_brightness',axes, **kwargs)

    def plot_data_image(self, axes, **kwargs):
        return self._plot_data_multi(axes, **kwargs)

    def plot_model_image(self, axes, **kwargs):
        return self._plot_lens_model_multi('plot_model_image', axes, **kwargs)

    def plot_model_residuals(self, axes, **kwargs):
        return self._plot_lens_model_multi('plot_model_residuals', axes, **kwargs)

    def plot_convergence(self, axes, **kwargs):
        return self._plot_lens_model_multi('plot_convergence', axes, **kwargs)

    def plot_magnification(self, axes, **kwargs):
        return self._plot_lens_model_multi('plot_magnification', axes, **kwargs)

    def plot_convergence_diff(self, axes, *args, **kwargs):
        return self._plot_lens_model_multi('plot_convergence_diff', axes, *args, **kwargs)

    def plot_magnification_diff(self, axes, *args, **kwargs):
        return self._plot_lens_model_multi('plot_magnification_diff', axes, *args, **kwargs)

    def _plot_light_multi(self, method_name, axes, **kwargs):
        assert len(axes) == self.num_models, "Inconsistent number of subplot axes"
        kwargs_ = copy.deepcopy(kwargs)
        image_list = []
        for i, (ax, plotter) in enumerate(zip(axes, self.plotter_list)):
            if ax is None:
                continue
            if 'kwargs_light' in kwargs:
                kwargs_['kwargs_light'] = {k: v[i] for k, v in kwargs['kwargs_light'].items()}
            if 'kwargs_lens_mass' in kwargs:  # used for over-plotting caustics
                kwargs_['kwargs_lens_mass'] = {k: v[i] for k, v in kwargs['kwargs_lens_mass'].items()}
            image = getattr(plotter, method_name)(ax, **kwargs_)
            image_list.append(image)
        return image_list

    def _plot_mass_multi(self, method_name, axes, **kwargs):
        assert len(axes) == self.num_models, "Inconsistent number of subplot axes"
        kwargs_ = copy.deepcopy(kwargs)
        image_list = []
        for i, (ax, plotter) in enumerate(zip(axes, self.plotter_list)):
            if ax is None:
                continue
            if 'kwargs_lens_mass' in kwargs:
                kwargs_['kwargs_lens_mass'] = {k: v[i] for k, v in kwargs['kwargs_lens_mass'].items()}
            image = getattr(plotter, method_name)(ax, **kwargs_)
            image_list.append(image)
        return image_list

    def _plot_lens_model_multi(self, method_name, axes, *args, **kwargs):
        assert len(axes) == self.num_models, "Inconsistent number of subplot axes"
        kwargs_ = copy.deepcopy(kwargs)
        image_list = []
        for i, (ax, plotter) in enumerate(zip(axes, self.plotter_list)):
            if ax is None:
                continue
            if 'kwargs_source' in kwargs:
                kwargs_['kwargs_source'] = {k: v[i] for k, v in kwargs['kwargs_source'].items()}
            if 'kwargs_lens_mass' in kwargs:
                kwargs_['kwargs_lens_mass'] = {k: v[i] for k, v in kwargs['kwargs_lens_mass'].items()}
            image = getattr(plotter, method_name)(ax, *args, **kwargs_)
            image_list.append(image)
        return image_list

    def _plot_data_multi(self, axes, **kwargs):
        assert len(axes) == self.num_models, "Inconsistent number of subplot axes"
        image_list = []
        for i, (ax, plotter) in enumerate(zip(axes, self.plotter_list)):
            if ax is None:
                continue
            image = getattr(plotter, 'plot_data_image')(ax, **kwargs)
            image_list.append(image)
        return image_list


class ParametersPlotter(object):
    """Handles plot of analytical models in a comparative way

    Parameters
    ----------
    parameter_id_list : array
        A list of parameter unique ids obtained from lensing entities. Their order determines the order of the plot panels.
    coolest_objects : array
        A list of coolest objects that have a chain file associated to them.
    coolest_directories : array
        A list of paths matching the coolest files in 'chain_objs'.
    coolest_names : array, optional
        A list of labels for the coolest models in the 'chain_objs' list. Must have the same order as 'chain_objs'.
    ref_coolest_objects : array, optional
        A list of coolest objects that will be used as point estimates.
    ref_coolest_directories : array
        A list of paths matching the coolest files in 'point_estimate_objs'.
    ref_coolest_names : array, optional
        A list of labels for the models in the 'point_estimate_objs' list. Must have the same order as 'point_estimate_objs'.
    posterior_bool_list : list, optional
        List of bool to toggle errorbars on point-estimate values
    colors : list, optional
        List of pyplot color names to associate to each coolest model.
    linestyles : list, optional
        List of pyplot linesyles to associate to each coolest model.
    add_multivariate_margin_samples : bool, optional
        If True, will append to the list of compared models
        a new chain that is resampled from the multi-variate normal distribution,
        where its covariance matrix is computed from the marginalization of
        all samples from all models. By default False. 
    num_samples_per_model_margin : int, optional
        Number of samples to (randomly) draw from each model samples to concatenate
        before estimating the multi-variate normal marginalization.
    """

    np.random.seed(598237)  # fix the random seed for reproducibility
    
    def __init__(self, parameter_id_list, coolest_objects, coolest_directories=None, coolest_names=None,
                 ref_coolest_objects=None, ref_coolest_directories=None, ref_coolest_names=None,
                 posterior_bool_list=None, colors=None, linestyles=None,
                 add_multivariate_margin_samples=False, num_samples_per_model_margin=5_000):
        self.parameter_id_list = parameter_id_list
        self.coolest_objects = coolest_objects
        self.coolest_directories = coolest_directories
        if coolest_names is None:
            coolest_names = ["Model "+str(i) for i in range(len(coolest_objects))]
        self.coolest_names = coolest_names
        self.ref_coolest_objects = ref_coolest_objects
        self.ref_coolest_directories = ref_coolest_directories
        self.ref_coolest_names = ref_coolest_names
        self.ref_file_names = ref_coolest_names

        self.num_models = len(self.coolest_objects)
        self.num_params = len(self.parameter_id_list)
        if colors is None:
            colors = plt.cm.turbo(np.linspace(0.1, 0.9, self.num_models))
        self.colors = colors
        if linestyles is None:
            linestyles = ['-']*self.num_models
        self.linestyles = linestyles
        self.ref_linestyles = ['--', ':', '-.', '-']
        self.ref_markers = ['s', '^', 'o', '*']

        self._add_margin_samples = add_multivariate_margin_samples
        self._ns_per_model_margin = num_samples_per_model_margin
        self._color_margin = 'black'
        self._label_margin = "Combined"

        # self.posterior_bool_list = posterior_bool_list
        # self.param_lens, self.param_source = util.split_lens_source_params(
        #     self.coolest_objects, self.coolest_names, lens_light=False)

    def init_getdist(self, shift_sample_list=None, settings_mcsamples=None,
                     add_multivariate_margin_samples=False):
        """Initializes the getdist plotter.

        Parameters
        ----------
        shift_sample_list : dict
            Dictionary keyed by parameter ID to apply a uniform additive shift to
            all samples of that parameters posterior distribution.
        settings_mcsamples : dict, optional
            Keyword arguments passed as the `settings` argument of getdist.MCSamples, by default None

        Raises
        ------
        ValueError
            If the csv file containing samples is is not coma (,) separated.
        """
        chains.print_load_details = False # Just to silence messages
        parameter_id_set = set(self.parameter_id_list)

        if shift_sample_list is None:
            shift_sample_list = [None]*self.num_models
        
        # Get the values of the point_estimates
        point_estimates = []
        if self.ref_coolest_objects is not None:
            for coolest_obj in self.ref_coolest_objects:
                values = []
                for par in self.parameter_id_list:
                    param = coolest_obj.lensing_entities.get_parameter_from_id(par)
                    val = param.point_estimate.value
                    if val is None:
                        values.append(None)
                    else:
                        values.append(val)
                point_estimates.append(values)

        mcsamples = []
        samples_margin, weights_margin = None, None
        mysample_margin = None
        for i in range(self.num_models):
            chain_file = os.path.join(self.coolest_directories[i],self.coolest_objects[i].meta["chain_file_name"]) # Here get the chain file path for each coolest object

            # Each chain file can have a different number of free parameters
            f = open(chain_file)
            header = f.readline()
            f.close()

            if ';' in header:
                raise ValueError("Columns must be coma-separated (no semi-colon) in chain file.")

            chain_file_headers = header.split(',')
            num_cols = len(chain_file_headers)
            chain_file_headers.pop() # Remove the last column name that is the probability weights
            chain_file_headers_set = set(chain_file_headers)
            
            # Check that the given parameters are a subset of those in the chain file
            assert parameter_id_set.issubset(chain_file_headers_set), "Not all given parameters are free parameters for model %d (not in the chain file: %s)!" % (i,chain_file)

            # Set the labels for the parameters in the chain file
            labels = []
            for par_id in self.parameter_id_list:
                param = self.coolest_objects[i].lensing_entities.get_parameter_from_id(par_id)
                labels.append(param.latex_str.strip('$'))

            # Read parameter values and probability weights
            column_indices = [chain_file_headers.index(par_id) for par_id in self.parameter_id_list]
            columns_to_read = sorted(column_indices) + [num_cols-1]  # add last one for probability weights
            samples = pd.read_csv(chain_file, usecols=columns_to_read, delimiter=',')
        
            # Re-order columns to match self.parameter_id_list and labels
            sample_par_values = np.array(samples[self.parameter_id_list])

            # If needed, shift samples by a constant
            if shift_sample_list[i] is not None:
                for param_id, value in shift_sample_list[i].items():
                    sample_par_values[:, self.parameter_id_list.index(param_id)] += value
                    logging.info(f"posterior for parameter '{param_id}' from model '{self.coolest_names[i]}' "
                                 f"has been shifted by {value}.")

            # Clean-up the probability weights
            mypost = np.array(samples['probability_weights'])
            min_non_zero = np.min(mypost[np.nonzero(mypost)])
            sample_prob_weight = np.where(mypost<min_non_zero, min_non_zero, mypost)
            #sample_prob_weight = mypost

            # Create MCSamples object
            mysample = MCSamples(samples=sample_par_values, names=self.parameter_id_list,
                                 labels=labels, settings=settings_mcsamples)
            mysample.reweightAddingLogLikes(-np.log(sample_prob_weight))
            mcsamples.append(mysample)

            # if required, aggregate the samples in a "marginalized" posterior
            if self._add_margin_samples:
                if i == 0:
                    mysample_margin = copy.deepcopy(mysample)
                else:
                    # combine the sample such that the probability mass of each set of samples is the same
                    mysample_margin = mysample_margin.getCombinedSamplesWithSamples(mysample, sample_weights=(1, 1))
        
        if self._add_margin_samples:
            mcsamples.append(mysample_margin)

        self._mcsamples = mcsamples
        self.ref_values = point_estimates
        self.ref_values_markers = [dict(zip(self.parameter_id_list, values)) for values in self.ref_values]

    def get_mcsamples_getdist(self, with_margin=False):
        if not self._add_margin_samples or with_margin:
            return self._mcsamples
        else:
            return self._mcsamples[:-1]
    
    def get_margin_mcsamples_getdist(self):
        if not self._add_margin_samples:
            return None
        else:
            return self._mcsamples[-1]
    
    def plot_triangle_getdist(self, filled_contours=True, angles_range=None, 
                              linewidth_hist=2, linewidth_cont=2, linewidth_margin=4,
                              marker_linewidth=2, marker_size=15, 
                              axes_labelsize=None, legend_fontsize=None,
                              **subplot_kwargs):
        """Corner array of subplots using getdist.triangle_plot method.

        Parameters
        ----------
        subplot_size : int, optional
            Size of the getdist plot, by default 1
        filled_contours : bool, optional
            Wether or not to fill the 2D contours, by default True
        angles_range : _type_, optional
            Restrict the range of angle (containing 'phi' in their name) parameters, by default None
        linewidth_hist : int, optional
            Line width for 1D histograms, by default 2
        linewidth_cont : int, optional
            Line width for 2D contours, by default 1
        marker_size : int, optional
            Size of the reference (scatter) markers on 2D contours plots, by default 15

        Returns
        -------
        GetDistPlotter
            Instance of GetDistPlotter corresponding to the figure
        """
        line_args, contour_lws, contour_ls, colors, legend_labels \
            = self._prepare_getdist_plot(linewidth_hist, 
                                         lw_cont=linewidth_cont, 
                                         lw_margin=linewidth_margin)
        
        filled_contours = [filled_contours]*len(self._mcsamples)
        alphas = [1]*len(self._mcsamples)
        if self._add_margin_samples:
            filled_contours[-1] = True
            # alphas[-1] = 0.7
    
        # Make the plot
        g = plots.get_subplot_plotter(**subplot_kwargs)
        if legend_fontsize is not None:
            g.settings.legend_fontsize = legend_fontsize 
        if axes_labelsize is not None:
            g.settings.axes_labelsize = axes_labelsize 
        g.triangle_plot(
            self._mcsamples,
            params=self.parameter_id_list,
            legend_labels=legend_labels,
            filled=filled_contours,
            colors=colors,
            line_args=line_args,   # TODO: issue that linewidth settings in line_args are being overwritten by contour_lws
            contour_colors=self.colors,
            contour_lws=contour_lws,
            contour_ls=contour_ls,
            alphas=alphas,
        )
        
        # Add marker lines and points
        for k in range(0, len(self.ref_values)):
            g.add_param_markers(self.ref_values_markers[k], color='black', ls=self.ref_linestyles[k], 
                                lw=marker_linewidth)
            for i in range(0,self.num_params):
                val_x = self.ref_values[k][i]
                for j in range(i+1,self.num_params):
                    val_y = self.ref_values[k][j]
                    if val_x is not None and val_y is not None:
                        g.subplots[j,i].scatter(val_x, val_y, s=marker_size, facecolors='black',
                                                color='black', marker=self.ref_markers[k])


        # Set default ranges for angles
        if angles_range is None:
            angles_range = (-90, 90)
        for i in range(0, len(self.parameter_id_list)):
            dum = self.parameter_id_list[i].split('-')
            name = dum[-1]
            if name in ['phi','phi_ext']:
                xlim = g.subplots[i,i].get_xlim()
                #print(xlim)
            
                if xlim[0] < -90:
                    for ax in g.subplots[i:,i]:
                        ax.set_xlim(left=angles_range[0])
                    for ax in g.subplots[i,:i]:
                        ax.set_ylim(bottom=angles_range[0])
                if xlim[1] > 90:
                    for ax in g.subplots[i:,i]:
                        ax.set_xlim(right=angles_range[1])
                    for ax in g.subplots[i,:i]:
                        ax.set_ylim(top=angles_range[1])
        return g
    
    def plot_rectangle_getdist(self, x_param_ids, y_param_ids, subplot_size=1, 
                               legend_ncol=None, legend_fontsize=None, 
                               filled_contours=True, linewidth=1,
                               marker_size=15, axes_labelsize=None, **subplot_kwargs):
        """Array of (2D contours) subplots using getdist.rectangle_plot method.

        Parameters
        ----------
        subplot_size : int, optional
            Size of the getdist plot, by default 1
        filled_contours : bool, optional
            Wether or not to fill the 2D contours, by default True
        linewidth : int, optional
            Line width for 2D contours, by default 1
        marker_size : int, optional
            Size of the reference (scatter) markers on 2D contours plots, by default 15
        legend_ncol : number of columns in the legend

        Returns
        -------
        GetDistPlotter
            Instance of GetDistPlotter corresponding to the figure
        """
        line_args, _, _, colors, legend_labels = self._prepare_getdist_plot(linewidth)
        
        if legend_ncol is None:
            legend_ncol = 3
        # Make the plot
        g = plots.get_subplot_plotter(**subplot_kwargs)
        if legend_fontsize is not None:
            g.settings.legend_fontsize = legend_fontsize
        if axes_labelsize is not None:
            g.settings.axes_labelsize = axes_labelsize
        g.rectangle_plot(x_param_ids, y_param_ids, roots=self._mcsamples,
                         filled=filled_contours,
                         colors=colors,
                         legend_ncol=legend_ncol,
                         legend_labels=legend_labels,
                         line_args=line_args, 
                         contour_colors=self.colors)
        for k in range(len(self.ref_values)):
            g.add_param_markers(self.ref_values_markers[k], color='black', ls=self.ref_linestyles[k], lw=linewidth)
            for j, key_x in enumerate(x_param_ids):
                val_x = self.ref_values_markers[k][key_x]
                for i, key_y in enumerate(y_param_ids):
                    val_y = self.ref_values_markers[k][key_y]
                    if val_x is not None and val_y is not None:
                        g.subplots[i, j].scatter(val_x,val_y,s=marker_size,facecolors='black',color='black',marker=self.ref_markers[k])
        return g
    
    def plot_1d_getdist(self, num_columns=None, legend_ncol=None, 
                        legend_fontsize=None, axes_labelsize=None,
                        linewidth=1, **subplot_kwargs):
        """Array of 1D histogram subplots using getdist.plots_1d method.

        Parameters
        ----------
        subplot_size : int, optional
            Size of the getdist plot, by default 1
        linewidth : int, optional
            Line width for 2D contours, by default 1
        marker_size : int, optional
            Size of the reference (scatter) markers on 2D contours plots, by default 15
        legend_ncol : int, optional
            number of columns in the legend
        num_columns : int, optional
            number of columns of the subplot array

        Returns
        -------
        GetDistPlotter
            Instance of GetDistPlotter corresponding to the figure
        """
        line_args, _, _, colors, legend_labels = self._prepare_getdist_plot(linewidth)

        if num_columns is None:
            num_columns = self.num_models//2+1
        if legend_ncol is None:
            legend_ncol = 3
        # Make the plot
        g = plots.get_subplot_plotter(**subplot_kwargs)
        if legend_fontsize is not None:
            g.settings.legend_fontsize = legend_fontsize
        if axes_labelsize is not None:
            g.settings.axes_labelsize = axes_labelsize
        g.plots_1d(self._mcsamples,
                   params=self.parameter_id_list,
                   legend_labels=legend_labels,
                   colors=colors,
                   share_y=True,
                   line_args=line_args,
                   nx=num_columns, legend_ncol=legend_ncol,
        )
        for k in range(len(self.ref_values)):
            g.add_param_markers(self.ref_values_markers[k], color='black', ls=self.ref_linestyles[k], lw=linewidth)
        # for k in range(0, len(self.ref_values)):
        #     # Add vertical and horizontal lines
        #     for i in range(0, self.num_params):
        #         val = self.ref_values[k][i]
        #         ax = g.subplots.flatten()[i]
        #         if val is not None:
        #             ax.axvline(val, color='black', ls=self.ref_linestyles[k], alpha=1.0, lw=1)
        return g

    def plot_source(self, idx_file=0):
        f,ax = self.plotting_routine(self.param_source,idx_file)
        return f,ax
    
    def plot_lens(self, idx_file=0):
        f,ax = self.plotting_routine(self.param_lens,idx_file)
        return f,ax

    def plotting_routine(self, param_dict, idx_file=0):
        """
        plot the parameters

        INPUT
        -----
        param_dict: dict, organized dictonnary with all parameters results of the different files
        idx_file: int, chooses the file on which the choice of plotted parameters will be made
        (not very clear: basically in file 0 you may have a sersic fit and in file 1 sersic+shapelets. If you choose
         idx_file=0, you will plot the sersic results of both file. If you choose idx_file=1, you will plot all the
         sersic and shapelets parameters when available)
        """

        #find the numer of parameters to plot and define a nice looking figure
        number_param = len(param_dict[self.file_names[idx_file]])
        unused_figs = []
        if number_param <= 4:
            print('so few parameters not implemented yet')
        else:
            if number_param % 4 == 0:
                num_lines = int(number_param / 4.)
            else:
                num_lines = int(number_param / 4.) + 1

                for idx in range(3):
                    if (number_param + idx) % 4 != 0:
                        unused_figs.append(-idx - 1)
                    else:
                        break

        f, ax = plt.subplots(num_lines, 4, figsize=(4 * 3.5, 2.5 * num_lines))
        markers = ['*', '.', 's', '^','<','>','v','p','P','X','D','1','2','3','4','+']
        #may find a better way to define markers but right now, it is sufficient

        for j, file_name in enumerate(self.file_names):
            i = 0
            result = param_dict[file_name]
            for key in result.keys():
                idx_line = int(i / 4.)
                idx_col = i % 4
                p = result[key]
                m = markers[j]
                if self.posterior_bool_list[j]:
                    # UNCOMMENT IF NO ERROR BARS AVAILABLE ON SHEAR
                    #             if (j== 1) and (key=='SHEAR_0_gamma_ext' or key == 'SHEAR_0_phi_ext'):
                    #                 ax[idx_line,idx_col].plot(j,p['point_estimate'],marker=m,ls='',label=file_name)
                    #                 i+=1
                    #                 continue

                    #trick to plot correct error bars if close to the +180/-180 edge
                    if (key == 'SHEAR_0_phi_ext' or key == 'PEMD_0_phi'):
                        if p['percentile_16th'] > p['median']:
                            p['percentile_16th'] -= 180.
                        if p['percentile_84th'] < p['median']:
                            p['percentile_84th'] += 180.
                    ax[idx_line, idx_col].errorbar(j, p['median'], [[p['median'] - p['percentile_16th']],
                                                                    [p['percentile_84th'] - p['median']]],
                                                   marker=m, ls='', label=file_name)
                else:
                    ax[idx_line, idx_col].plot(j, p['point_estimate'], marker=m, ls='', label=file_name)

                if j == 0:
                    ax[idx_line, idx_col].get_xaxis().set_visible(False)
                    ax[idx_line, idx_col].set_ylabel(p['latex_str'], fontsize=12)
                    ax[idx_line, idx_col].tick_params(axis='y', labelsize=12)
                i += 1

        ax[0, 0].legend()
        for idx in unused_figs:
            ax[-1, idx].axis('off')
        plt.tight_layout()
        plt.show()
        return f, ax

    def _prepare_getdist_plot(self, lw, lw_cont=None, lw_margin=None):
        if lw_margin is None:
            lw_margin = lw + 2
        line_args = [{'ls': ls, 'lw': lw, 'color': c} for ls, c in zip(self.linestyles, self.colors)]
        lw_conts = [lw_cont]*self.num_models
        ls_conts = self.linestyles
        legend_labels = copy.deepcopy(self.coolest_names)
        colors = copy.deepcopy(self.colors)
        if self._add_margin_samples:
            line_args.append({'ls': '-.', 'lw': lw_margin, 'alpha': 0.8, 'color': self._color_margin})
            ls_conts.append('-.')
            if lw_cont is not None: lw_conts.append(lw_margin)
            legend_labels.append(self._label_margin)
            colors.append(self._color_margin)
        return line_args, lw_conts, ls_conts, colors, legend_labels

# def plot_corner(parameter_id_list, 
#                 chain_objs, chain_dirs, chain_names=None, 
#                 point_estimate_objs=None, point_estimate_dirs=None, point_estimate_names=None, 
#                 colors=None, labels=None, subplot_size=1, mc_samples_kwargs=None, 
#                 filled_contours=True, angles_range=None, shift_sample_list=None):
#     """
#     Adding this as just a function for the moment.
#     Takes a list of COOLEST files as input, which must have a chain file associated to them, and returns a corner plot.

#     Parameters
#     ----------
#     parameter_id_list : array
#         A list of parameter unique ids obtained from lensing entities. Their order determines the order of the plot panels.
#     chain_objs : array
#         A list of coolest objects that have a chain file associated to them.
#     chain_dirs : array
#         A list of paths matching the coolest files in 'chain_objs'.
#     chain_names : array, optional
#         A list of labels for the coolest models in the 'chain_objs' list. Must have the same order as 'chain_objs'.
#     point_estimate_objs : array, optional
#         A list of coolest objects that will be used as point estimates.
#     point_estimate_dirs : array
#         A list of paths matching the coolest files in 'point_estimate_objs'.
#     point_estimate_names : array, optional
#         A list of labels for the models in the 'point_estimate_objs' list. Must have the same order as 'point_estimate_objs'.
#     labels : dict, optional
#         A dictionary matching the parameter_id_list entries to some human-readable labels.

#     Returns
#     -------
#     An image
#     """

#     chains.print_load_details = False # Just to silence messages
#     parameter_id_set = set(parameter_id_list)
#     Npars = len(parameter_id_list)
#     Nobjs = len(chain_objs)
    
#     # Set the chain names
#     if chain_names is None:
#         chain_names = ["chain "+str(i) for i in range(Nobjs)]
    
#     if shift_sample_list is None:
#         shift_sample_list = [None]*Nobjs
    
#     # Get the values of the point_estimates
#     point_estimates = []
#     if point_estimate_objs is not None:
#         for coolest_obj in point_estimate_objs:
#             values = []
#             for par in parameter_id_list:
#                 param = coolest_obj.lensing_entities.get_parameter_from_id(par)
#                 val = param.point_estimate.value
#                 if val is None:
#                     values.append(None)
#                 else:
#                     values.append(val)
#             point_estimates.append(values)


            
#     mcsamples = []
#     for i in range(Nobjs):
#         chain_file = os.path.join(chain_dirs[i],chain_objs[i].meta["chain_file_name"]) # Here get the chain file path for each coolest object

#         # Each chain file can have a different number of free parameters
#         f = open(chain_file)
#         header = f.readline()
#         f.close()

#         if ';' in header:
#             raise ValueError("Columns must be coma-separated (no semi-colon) in chain file.")

#         chain_file_headers = header.split(',')
#         num_cols = len(chain_file_headers)
#         chain_file_headers.pop() # Remove the last column name that is the probability weights
#         chain_file_headers_set = set(chain_file_headers)
        
#         # Check that the given parameters are a subset of those in the chain file
#         assert parameter_id_set.issubset(chain_file_headers_set), "Not all given parameters are free parameters for model %d (not in the chain file: %s)!" % (i,chain_file)

#         # Set the labels for the parameters in the chain file
#         par_labels = []
#         if labels is None:
#             labels = {}
#         for par_id in parameter_id_list:
#             if labels.get(par_id, None) is None:
#                 param = coolest_obj.lensing_entities.get_parameter_from_id(par_id)
#                 par_labels.append(param.latex_str.strip('$'))
#             else:
#                 par_labels.append(labels[par_id])
                    
#         # Read parameter values and probability weights
#         column_indices = [chain_file_headers.index(par_id) for par_id in parameter_id_list]
#         columns_to_read = sorted(column_indices) + [num_cols-1]  # add last one for probability weights
#         samples = pd.read_csv(chain_file, usecols=columns_to_read, delimiter=',')
    
#         # Re-order columnds to match parameter_id_list and par_labels
#         sample_par_values = np.array(samples[parameter_id_list])

#         # If needed, shift samples by a constant
#         if shift_sample_list[i] is not None:
#             for param_id, value in shift_sample_list[i].items():
#                 sample_par_values[:, parameter_id_list.index(param_id)] += value
#                 print(f"INFO: posterior for parameter '{param_id}' from model '{chain_names[i]}' "
#                       f"has been shifted by {value}.")

#         # Clean-up the probability weights
#         mypost = np.array(samples['probability_weights'])
#         min_non_zero = np.min(mypost[np.nonzero(mypost)])
#         sample_prob_weight = np.where(mypost<min_non_zero,min_non_zero,mypost)
#         #sample_prob_weight = mypost

#         # Create MCSamples object
#         mysample = MCSamples(samples=sample_par_values,names=parameter_id_list,labels=par_labels,settings=mc_samples_kwargs)
#         mysample.reweightAddingLogLikes(-np.log(sample_prob_weight))
#         mcsamples.append(mysample)


        
#     # Make the plot
#     image = plots.getSubplotPlotter(subplot_size=subplot_size)    
#     image.triangle_plot(mcsamples,
#                         params=parameter_id_list,
#                         legend_labels=chain_names,
#                         filled=filled_contours,
#                         colors=colors,
#                         line_args=[{'ls':'-', 'lw': 2, 'color': c} for c in colors], 
#                         contour_colors=colors)


#     my_linestyles = ['solid','dotted','dashed','dashdot']
#     my_markers    = ['s','^','o','star']

#     for k in range(0,len(point_estimates)):
#         # Add vertical and horizontal lines
#         for i in range(0,Npars):
#             val = point_estimates[k][i]
#             if val is not None:
#                 for ax in image.subplots[i:,i]:
#                     ax.axvline(val,color='black',ls=my_linestyles[k],alpha=1.0,lw=1)
#                 for ax in image.subplots[i,:i]:
#                     ax.axhline(val,color='black',ls=my_linestyles[k],alpha=1.0,lw=1)

#         # Add points
#         for i in range(0,Npars):
#             val_x = point_estimates[k][i]
#             for j in range(i+1,Npars):
#                 val_y = point_estimates[k][j]
#                 if val_x is not None and val_y is not None:
#                     image.subplots[j,i].scatter(val_x,val_y,s=10,facecolors='black',color='black',marker=my_markers[k])
#                 else:
#                     pass    


#     # Set default ranges for angles
#     if angles_range is None:
#         angles_range = (-90, 90)
#     for i in range(0,len(parameter_id_list)):
#         dum = parameter_id_list[i].split('-')
#         name = dum[-1]
#         if name in ['phi','phi_ext']:
#             xlim = image.subplots[i,i].get_xlim()
#             #print(xlim)
        
#             if xlim[0] < -90:
#                 for ax in image.subplots[i:,i]:
#                     ax.set_xlim(left=angles_range[0])
#                 for ax in image.subplots[i,:i]:
#                     ax.set_ylim(bottom=angles_range[0])
#             if xlim[1] > 90:
#                 for ax in image.subplots[i:,i]:
#                     ax.set_xlim(right=angles_range[1])
#                 for ax in image.subplots[i,:i]:
#                     ax.set_ylim(top=angles_range[1])

            
#     return image



__author__ = 'aymgal'

from typing import Tuple

from coolest.template.classes.profile_list import ProfileList
from coolest.template.classes.profiles import mass as mass_profiles_module
from coolest.template.classes.profiles import light as light_profiles_module


class MassModel(ProfileList):
    """Describes a mass model of a lensing entity as a list of (mass) profiles
    
    Parameters
    ----------
    *profile_names : str
        Names of the mass profiles, following corresponding class names in
        the coolest.template.classes.profiles.mass submodule.
    """

    def __init__(self, 
                 *profile_names: Tuple[str]) -> None:
        super().__init__(mass_profiles_module, *profile_names)
        

class LightModel(ProfileList):
    """Describes a light model of a lensing entity as a list of (light) profiles

    Parameters
    ----------
    *profile_names : str
        Names of the light profiles, following corresponding class names in
        the coolest.template.classes.profiles.light submodule.
    """

    def __init__(self, 
                 *profile_names: Tuple[str]) -> None:
        super().__init__(light_profiles_module, *profile_names)
